import numpy as np
import cv2
from zebrazoom.code.preprocessImage import preprocessImage, preprocessBackgroundImage
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading


def getBackground(videoPath, hyperparameters):
  
  cap   = zzVideoReading.VideoCapture(videoPath)
  max_l = int(cap.get(7))

  backCalculationStep = hyperparameters["backCalculationStep"]
  if ("firstFrameForBackExtract" in hyperparameters) and ("lastFrameForBackExtract" in hyperparameters) and hyperparameters["firstFrameForBackExtract"] != -1 and hyperparameters["lastFrameForBackExtract"] != -1:
    firstFrame = hyperparameters["firstFrameForBackExtract"]
    lastFrame  = hyperparameters["lastFrameForBackExtract"]
  else:
    if ("firstFrame" in hyperparameters) and (hyperparameters["backgroundExtractionForceUseAllVideoFrames"] == 0):
      firstFrame = hyperparameters["firstFrame"]
    else:
      firstFrame = 1
    if ("lastFrame" in hyperparameters) and (hyperparameters["backgroundExtractionForceUseAllVideoFrames"] == 0):
      lastFrame  = hyperparameters["lastFrame"]
    else:
      lastFrame  = max_l - 10
  
  debugExtractBack    = hyperparameters["debugExtractBack"]
  
  if backCalculationStep == -1:
    backCalculationStep = int((lastFrame - firstFrame) / hyperparameters["nbImagesForBackgroundCalculation"])
    if backCalculationStep <= 1:
      backCalculationStep = 1
  
  ret, back = cap.read()
  
  if hyperparameters["useFirstFrameAsBackground"]:
    if hyperparameters["invertBlackWhiteOnImages"]:
      back = 255 - back
    back = cv2.cvtColor(back, cv2.COLOR_BGR2GRAY)
    if hyperparameters["backgroundPreProcessMethod"]:
      back = preprocessBackgroundImage(back, hyperparameters)
    if debugExtractBack:
      from PyQt5.QtCore import Qt
      from PyQt5.QtWidgets import QLabel, QVBoxLayout
      import zebrazoom.code.util as util
      label = QLabel()
      label.setMinimumSize(1, 1)
      layout = QVBoxLayout()
      layout.addWidget(label, alignment=Qt.AlignmentFlag.AlignCenter)
      timeout = 3000 if hyperparameters["exitAfterBackgroundExtraction"] else None
      util.showDialog(layout, title="Background Extracted", labelInfo=(back, label), timeout=timeout)
    cap.release()
    print("Background Extracted from first frame")
    return back
  
  if ret and hyperparameters["invertBlackWhiteOnImages"]:
    back = 255 - back
  back = cv2.cvtColor(back, cv2.COLOR_BGR2GRAY)
  if hyperparameters["backgroundExtractionWithOnlyTwoFrames"] == 0:
    for k in range(firstFrame,lastFrame):
      if (k % backCalculationStep == 0):
        cap.set(1, k)
        ret, frame = cap.read()
        if ret and hyperparameters["invertBlackWhiteOnImages"]:
          frame = 255 - frame
        if debugExtractBack:
          print(k)
        if ret:
          frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
          if hyperparameters["extractBackWhiteBackground"]:
            back = cv2.max(frame, back)
          else:
            back = cv2.min(frame, back)
        else:
          print("couldn't use the frame", k, "for the background extraction")
  else:
    maxDiff    = 0
    indMaxDiff = firstFrame
    for k in range(firstFrame,lastFrame):
      if (k % backCalculationStep == 0):
        cap.set(1, k)
        ret, frame = cap.read()
        if ret:
          if hyperparameters["invertBlackWhiteOnImages"]:
            frame = 255 - frame
          frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
          diff  = np.sum(np.abs(frame - back))
          if diff > maxDiff:
            maxDiff    = diff
            indMaxDiff = k
    cap.set(1, indMaxDiff)
    ret, frame = cap.read()
    if ret:
      if hyperparameters["invertBlackWhiteOnImages"]:
        frame = 255 - frame
      frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
      if hyperparameters["extractBackWhiteBackground"]:
        back = cv2.max(frame, back)
      else:
        back = cv2.min(frame, back)
    else:
      print("couldn't use the frame", k, "for the background extraction")
  
  if hyperparameters["backgroundPreProcessMethod"]:
    back = preprocessBackgroundImage(back, hyperparameters)
  
  if hyperparameters["checkThatMovementOccurInVideo"]:
    cap.set(1, firstFrame)
    ret, frame = cap.read()
    if ret:
      if hyperparameters["invertBlackWhiteOnImages"]:
        frame = 255 - frame
      frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
      if hyperparameters["imagePreProcessMethod"]:
        frame = preprocessBackgroundImage(frame, hyperparameters)
      if type(frame[0][0]) == np.ndarray:
        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
      putToWhite = (frame.astype('int32') >= (back.astype('int32')-hyperparameters["minPixelDiffForBackExtract"]))
      frame[putToWhite] = 255
    firstImage = frame
    maxDiff    = 0
    indMaxDiff = firstFrame
    for k in range(firstFrame,lastFrame):
      if (k % backCalculationStep == 0):
        cap.set(1, k)
        ret, frame = cap.read()
        if ret:
          if hyperparameters["invertBlackWhiteOnImages"]:
            frame = 255 - frame
          frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
          if hyperparameters["imagePreProcessMethod"]:
            frame = preprocessBackgroundImage(frame, hyperparameters)
          if type(frame[0][0]) == np.ndarray:
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
          putToWhite = (frame.astype('int32') >= (back.astype('int32')-hyperparameters["minPixelDiffForBackExtract"]))
          frame[putToWhite] = 255
          if hyperparameters["checkThatMovementOccurInVideoMedianFilterWindow"]:
            diff  = np.sum(cv2.medianBlur(np.abs(frame - firstImage), hyperparameters["checkThatMovementOccurInVideoMedianFilterWindow"]))
          else:
            diff  = np.sum(np.abs(frame - firstImage))
          if diff > maxDiff:
            maxDiff    = diff
            indMaxDiff = k
    print("checkThatMovementOccurInVideo: max difference is:", maxDiff)
    if maxDiff < hyperparameters["checkThatMovementOccurInVideo"]:
      back[:, :] = 0 # TODO: tracking should NOT RUN after background is set to 0 as it is here
  
  if hyperparameters["setBackgroundToImageMedian"]:
    back[:, :] = np.median(back)
  
  if debugExtractBack:
    from PyQt5.QtCore import Qt
    from PyQt5.QtWidgets import QLabel, QVBoxLayout

    import zebrazoom.code.util as util

    label = QLabel()
    label.setMinimumSize(1, 1)
    layout = QVBoxLayout()
    layout.addWidget(label, alignment=Qt.AlignmentFlag.AlignCenter)
    timeout = 3000 if hyperparameters["exitAfterBackgroundExtraction"] else None
    util.showDialog(layout, title="Background Extracted", labelInfo=(back, label), timeout=timeout)
  cap.release()
  
  print("Background Extracted")
  if hyperparameters["popUpAlgoFollow"]:
    import zebrazoom.code.popUpAlgoFollow as popUpAlgoFollow
    popUpAlgoFollow.prepend("Background Extracted")
  
  return back



from scipy.optimize import linear_sum_assignment
import numpy as np
import math

def findOptimalIdCorrespondance(trackingHeadTailAllAnimalsList, wellNumber, i,  firstFrame):
  
  if i > firstFrame:
    
    costMatrix = np.zeros((len(trackingHeadTailAllAnimalsList[wellNumber]), len(trackingHeadTailAllAnimalsList[wellNumber])))
    
    for animalIdPrev in range(0, len(trackingHeadTailAllAnimalsList[wellNumber])):
      for animalIdCur in range(0, len(trackingHeadTailAllAnimalsList[wellNumber])):
        coordPrevX = trackingHeadTailAllAnimalsList[wellNumber][animalIdPrev, i-firstFrame-1][0][0]
        coordPrevY = trackingHeadTailAllAnimalsList[wellNumber][animalIdPrev, i-firstFrame-1][0][1]         
        coordCurX  = trackingHeadTailAllAnimalsList[wellNumber][animalIdCur,  i-firstFrame][0][0]
        coordCurY  = trackingHeadTailAllAnimalsList[wellNumber][animalIdCur,  i-firstFrame][0][1]
        # TO DO: add some very high cost for (0, 0) coordinates
        costMatrix[animalIdPrev, animalIdCur] = math.sqrt((coordCurX - coordPrevX)**2 + (coordCurY - coordPrevY)**2)
    
    row_ind, col_ind = linear_sum_assignment(costMatrix)
    
    return col_ind
    
  else:
    
    return np.array([k for k in range(0, len(trackingHeadTailAllAnimalsList[wellNumber]))])


def switchIdentities(correspondance, trackingHeadTailAllAnimalsList, trackingHeadingAllAnimalsList, wellNumber, i, firstFrame):
  
  trackingHeadTailAllAnimalsListWellNumberOriginal = trackingHeadTailAllAnimalsList[wellNumber][:, i-firstFrame].copy()
  trackingHeadingAllAnimalsListWellNumberOriginal  = trackingHeadingAllAnimalsList[wellNumber][:, i-firstFrame].copy()
  
  for previousId, newId in enumerate(correspondance):
    trackingHeadTailAllAnimalsList[wellNumber][previousId, i-firstFrame] = trackingHeadTailAllAnimalsListWellNumberOriginal[newId]

  for previousId, newId in enumerate(correspondance):
    trackingHeadingAllAnimalsList[wellNumber][previousId, i-firstFrame] = trackingHeadingAllAnimalsListWellNumberOriginal[newId]
  
  return [trackingHeadTailAllAnimalsList, trackingHeadingAllAnimalsList]




from zebrazoom.code.trackingFolder.tracking import tracking
from zebrazoom.code.extractParameters import extractParameters
from zebrazoom.code.trackingFolder.headTrackingHeadingCalculationFolder.headTrackingHeadingCalculation import headTrackingHeadingCalculation
from zebrazoom.code.trackingFolder.tailTracking import tailTracking
from zebrazoom.code.trackingFolder.eyeTracking.eyeTracking import eyeTracking
from zebrazoom.code.trackingFolder.postProcessMultipleTrajectories import postProcessMultipleTrajectories
from zebrazoom.code.trackingFolder.getImages import getImages
from zebrazoom.code.trackingFolder.debugTracking import debugTracking
from zebrazoom.code.adjustHyperparameters import adjustFreelySwimTrackingParams
from zebrazoom.code.preprocessImage import preprocessImage
import multiprocessing as mp
from multiprocessing import Process
import cv2
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
import numpy as np
import math
import time
import zebrazoom.code.util as util
import queue

from zebrazoom.code.trackingFolder.tailTrackingFunctionsFolder.tailTrackingExtremityDetectFolder.getMidline import getMidline
from zebrazoom.code.trackingFolder.tailTrackingFunctionsFolder.tailTrackingExtremityDetectFolder.findTailExtremeteFolder.findTailExtremete import findTailExtremete
from zebrazoom.code.trackingFolder.tailTrackingFunctionsFolder.tailTrackingExtremityDetectFolder.findTheTwoSides import findTheTwoSides
from zebrazoom.code.trackingFolder.tailTrackingFunctionsFolder.tailTrackingExtremityDetectFolder.findBodyContour import findBodyContour
from zebrazoom.code.trackingFolder.tailTrackingFunctionsFolder.tailTrackingExtremityDetectFolder.Rotate import Rotate
from zebrazoom.code.trackingFolder.trackingFunctions import calculateAngle

from zebrazoom.code.trackingFolder.refactoredCode2022.findCenterByIterativelyDilating import findCenterByIterativelyDilating
from zebrazoom.code.trackingFolder.refactoredCode2022.headingCompute import computeHeading2
from zebrazoom.code.trackingFolder.refactoredCode2022.findTheTwoSides2 import findTheTwoSides2
from zebrazoom.code.trackingFolder.refactoredCode2022.identitiesLinkage import findOptimalIdCorrespondance, switchIdentities

from zebrazoom.code.updateBackgroundAtInterval import updateBackgroundAtInterval

from zebrazoom.code.trackingFolder.refactoredCode2022.detectMovementWithRawVideoInsideTracking2 import detectMovementWithRawVideoInsideTracking2

def fasterMultiprocessing2(videoPath, background, wellPositions, output, hyperparameters, videoName):
  
  cap = zzVideoReading.VideoCapture(videoPath)
  if (cap.isOpened()== False): 
    print("Error opening video stream or file")
  frame_width  = int(cap.get(3))
  frame_height = int(cap.get(4))
  firstFrame = hyperparameters["firstFrame"]
  lastFrame  = hyperparameters["lastFrame"]
  nbTailPoints = hyperparameters["nbTailPoints"]
  
  trackingHeadTailAllAnimalsList = []
  trackingHeadingAllAnimalsList  = []
  # if hyperparameters["eyeTracking"]:
    # trackingEyesAllAnimalsList   = []
  # else:
    # trackingEyesAllAnimals = 0
  trackingDataList               = []
  
  if hyperparameters["detectMovementWithRawVideoInsideTracking"]:
    previousFrames   = queue.Queue(hyperparameters["frameGapComparision"])
    # previousXYCoords = queue.Queue(hyperparameters["frameGapComparision"])
    auDessusPerAnimalIdList = []
    for wellNumber in range(0, hyperparameters["nbWells"]):
      auDessusPerAnimalIdList.append([np.zeros((lastFrame-firstFrame+1, 1)) for nbAnimalsPerWell in range(0, hyperparameters["nbAnimalsPerWell"])])
  
  # if not(hyperparameters["nbAnimalsPerWell"] > 1) and not(hyperparameters["headEmbeded"]) and (hyperparameters["findHeadPositionByUserInput"] == 0) and (hyperparameters["takeTheHeadClosestToTheCenter"] == 0):
    # trackingProbabilityOfGoodDetectionList = []
  # else:
    # trackingProbabilityOfGoodDetectionList = 0
  # trackingProbabilityOfGoodDetectionList = []
  
  for wellNumber in range(0, hyperparameters["nbWells"]):
    trackingHeadTailAllAnimalsList.append(np.zeros((hyperparameters["nbAnimalsPerWell"], lastFrame-firstFrame+1, nbTailPoints, 2)))
    trackingHeadingAllAnimalsList.append(np.zeros((hyperparameters["nbAnimalsPerWell"], lastFrame-firstFrame+1)))
    # if hyperparameters["eyeTracking"]:
      # trackingEyesAllAnimalsList.append(np.zeros((hyperparameters["nbAnimalsPerWell"], lastFrame-firstFrame+1, 8)))
    # trackingProbabilityOfGoodDetectionList.append(np.zeros((hyperparameters["nbAnimalsPerWell"], lastFrame-firstFrame+1)))
  
  # if hyperparameters["backgroundSubtractorKNN"]:
    # fgbg = cv2.createBackgroundSubtractorKNN()
    # for i in range(0, min(lastFrame - 1, 500), int(min(lastFrame - 1, 500) / 10)):
      # cap.set(1, min(lastFrame - 1, 500) - i)
      # ret, frame = cap.read()
      # fgmask = fgbg.apply(frame)
    # cap.release()
    # cap = zzVideoReading.VideoCapture(videoPath)
  
  i = firstFrame
  
  if firstFrame:
    cap.set(1, firstFrame)
  
  widgets = None
  while (i < lastFrame + 1):
    
    if (hyperparameters["freqAlgoPosFollow"] != 0) and (i % hyperparameters["freqAlgoPosFollow"] == 0):
      print("Tracking: frame:",i)
      if hyperparameters["popUpAlgoFollow"]:
        from zebrazoom.code.popUpAlgoFollow import prepend

        prepend("Tracking: frame:" + str(i))
    
    if hyperparameters["debugTracking"]:
      print("frame:",i)
      
    ret, frame = cap.read()
    
    if ret:
      
      if hyperparameters["imagePreProcessMethod"]:
        frame = preprocessImage(frame, hyperparameters)
      
      # if hyperparameters["backgroundSubtractorKNN"]:
        # frame = fgbg.apply(frame)
        # frame = 255 - frame
      
      for wellNumber in range(0 if hyperparameters["onlyTrackThisOneWell"] == -1 else hyperparameters["onlyTrackThisOneWell"], hyperparameters["nbWells"] if hyperparameters["onlyTrackThisOneWell"] == -1 else hyperparameters["onlyTrackThisOneWell"] + 1):
        
        minPixelDiffForBackExtract = hyperparameters["minPixelDiffForBackExtract"]
        xtop = wellPositions[wellNumber]['topLeftX']
        ytop = wellPositions[wellNumber]['topLeftY']
        lenX = wellPositions[wellNumber]['lengthX']
        lenY = wellPositions[wellNumber]['lengthY']
        # if hyperparameters["backgroundSubtractorKNN"]:
          # grey = frame
        # else:
        grey = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        
        curFrame = grey[ytop:ytop+lenY, xtop:xtop+lenX].copy()
        initialCurFrame = curFrame.copy()
        # if not(hyperparameters["backgroundSubtractorKNN"]):
        back = background[ytop:ytop+lenY, xtop:xtop+lenX]
        putToWhite = ( curFrame.astype('int32') >= (back.astype('int32') - minPixelDiffForBackExtract) )
        curFrame[putToWhite] = 255
        # else:
          # hyperparameters["paramGaussianBlur"] = int(math.sqrt(cv2.countNonZero(255 - curFrame) / hyperparameters["nbAnimalsPerWell"]) / 2) * 2 + 1
        # if hyperparameters["paramGaussianBlur"]:
          # blur = cv2.GaussianBlur(curFrame, (hyperparameters["paramGaussianBlur"], hyperparameters["paramGaussianBlur"]),0)
        # else:
          # blur = curFrame
        headPositionFirstFrame = 0
        
        ret, thresh1 = cv2.threshold(curFrame.copy(), 254, 255, cv2.THRESH_BINARY)
        
        contours, hierarchy = cv2.findContours(thresh1,cv2.RETR_TREE,cv2.CHAIN_APPROX_SIMPLE)
        areas = np.array([cv2.contourArea(contour) for contour in contours])
        
        maxIndexes = []
        for numFish in range(0, hyperparameters["nbAnimalsPerWell"]):
          maxArea = -1
          maxInd  = -1
          for idx, area in enumerate(areas):
            if area > maxArea and area > hyperparameters["minAreaBody"] and area < hyperparameters["maxAreaBody"]:
              maxArea = area
              maxInd  = idx
          areas[maxInd] = -1
          if maxInd != -1:
            maxIndexes.append(maxInd)
        
        for animal_Id, idx in enumerate(maxIndexes):
          bodyContour = contours[idx]
          M = cv2.moments(bodyContour)
          if M['m00']:
            # x = int(M['m10']/M['m00'])
            # y = int(M['m01']/M['m00'])
            # headPosition = [x, y]
            headPosition = findCenterByIterativelyDilating(bodyContour.copy(), len(curFrame[0]), len(curFrame))
            
            trackingHeadTailAllAnimalsList[wellNumber][animal_Id, i-firstFrame][0][0] = headPosition[0]
            trackingHeadTailAllAnimalsList[wellNumber][animal_Id, i-firstFrame][0][1] = headPosition[1]
            
            heading = computeHeading2(bodyContour.copy(), len(curFrame[0]), len(curFrame), headPosition, hyperparameters)
            
            trackingHeadingAllAnimalsList[wellNumber][animal_Id, i-firstFrame] = heading
            
            if hyperparameters["trackTail"] == 1 :
              
              res = findTheTwoSides2(headPosition, bodyContour, curFrame, hyperparameters, heading)
              
              # Finding tail extremity
              rotatedContour = bodyContour.copy()
              rotatedContour = Rotate(rotatedContour,int(headPosition[0]),int(headPosition[1]),heading,curFrame)
              debugAdv = False
              
              [MostCurvyIndex, distance2] = findTailExtremete(rotatedContour, bodyContour, headPosition[0], int(res[0]), int(res[1]), debugAdv, curFrame, hyperparameters["tailExtremityMaxJugeDecreaseCoeff"], hyperparameters)
              
              # Getting Midline
              if hyperparameters["detectMouthInsteadOfHeadTwoSides"] == 0:
                tail = getMidline(int(res[0]), int(res[1]), int(MostCurvyIndex), bodyContour, curFrame, nbTailPoints-1, distance2, debugAdv, hyperparameters, nbTailPoints)
              else:
                tail = getMidline(int(res[0]), int(res[1]), int(MostCurvyIndex), bodyContour, curFrame, nbTailPoints, distance2, debugAdv, hyperparameters, nbTailPoints)
                tail = np.array([tail[0][1:len(tail[0])]])
              tail = np.insert(tail, 0, headPosition, axis=1)
              trackingHeadTailAllAnimalsList[wellNumber][animal_Id, i-firstFrame] = tail
        
        # Eye tracking for frame i
        # if hyperparameters["eyeTracking"]:
          # trackingEyesAllAnimalsList[wellNumber] = eyeTracking(animalId, i, firstFrame, frame, hyperparameters, thresh1, trackingHeadingAllAnimalsList[wellNumber], trackingHeadTailAllAnimalsList[wellNumber], trackingEyesAllAnimalsList[wellNumber])
        
        correspondance = findOptimalIdCorrespondance(trackingHeadTailAllAnimalsList, wellNumber,  i, firstFrame)
        
        [trackingHeadTailAllAnimalsList, trackingHeadingAllAnimalsList] = switchIdentities(correspondance, trackingHeadTailAllAnimalsList, trackingHeadingAllAnimalsList, wellNumber, i, firstFrame)
        
        debugTracking(nbTailPoints, i, firstFrame, trackingHeadTailAllAnimalsList[wellNumber], trackingHeadingAllAnimalsList[wellNumber], curFrame, hyperparameters)
        
        if hyperparameters["updateBackgroundAtInterval"]:
          background = updateBackgroundAtInterval(i, hyperparameters, background, wellPositions, wellNumber, initialCurFrame, firstFrame, trackingHeadTailAllAnimalsList[wellNumber], initialCurFrame)
        
        if hyperparameters["freqAlgoPosFollow"]:
          if i % hyperparameters["freqAlgoPosFollow"] == 0:
            print("Tracking at frame", i)
        
      if hyperparameters["detectMovementWithRawVideoInsideTracking"]:
        for numFish in range(0, hyperparameters["nbAnimalsPerWell"]):
          [auDessusPerAnimalIdList, previousFrames] = detectMovementWithRawVideoInsideTracking2(hyperparameters, trackingHeadTailAllAnimalsList, previousFrames, numFish, i, firstFrame, auDessusPerAnimalIdList, grey, wellPositions)
      
    # if hyperparameters["adjustFreelySwimTracking"] == 1:
      # i, widgets = adjustFreelySwimTrackingParams(nbTailPoints, i, firstFrame, trackingHeadTailAllAnimals, trackingHeadingAllAnimals, frame, frame2, hyperparameters, widgets)
    # else:

    i = i + 1
    
  for wellNumber in range(0 if hyperparameters["onlyTrackThisOneWell"] == -1 else hyperparameters["onlyTrackThisOneWell"], hyperparameters["nbWells"] if hyperparameters["onlyTrackThisOneWell"] == -1 else hyperparameters["onlyTrackThisOneWell"] + 1):
    
    # if hyperparameters["postProcessMultipleTrajectories"]:
      # [trackingHeadingAllAnimalsList[wellNumber], trackingHeadTailAllAnimalsList[wellNumber], trackingEyesAllAnimals] = postProcessMultipleTrajectories(trackingHeadingAllAnimalsList[wellNumber], trackingHeadTailAllAnimalsList[wellNumber], [], trackingProbabilityOfGoodDetectionList[wellNumber], hyperparameters, wellPositions)
    if hyperparameters["detectMovementWithRawVideoInsideTracking"]:
      trackingDataList.append([trackingHeadTailAllAnimalsList[wellNumber], trackingHeadingAllAnimalsList[wellNumber], [], 0, 0, auDessusPerAnimalIdList[wellNumber]])
    else:
      trackingDataList.append([trackingHeadTailAllAnimalsList[wellNumber], trackingHeadingAllAnimalsList[wellNumber], [], 0, 0])
  
  wellNumberBeginLoop = 0 if hyperparameters["onlyTrackThisOneWell"] == -1 else hyperparameters["onlyTrackThisOneWell"]
  for wellNumber in range(wellNumberBeginLoop, hyperparameters["nbWells"] if hyperparameters["onlyTrackThisOneWell"] == -1 else hyperparameters["onlyTrackThisOneWell"] + 1):
    parameters = extractParameters(trackingDataList[wellNumber-wellNumberBeginLoop], wellNumber, hyperparameters, videoPath, wellPositions, background)
    output.append([wellNumber,parameters,[]])
  
  return output



import math
import sys

import numpy as np

import cv2

from PyQt5.QtCore import pyqtSignal, Qt, QAbstractAnimation, QAbstractEventDispatcher, QEventLoop, QLine, QParallelAnimationGroup, QPoint, QPointF, QPropertyAnimation, QRectF, QSize, QSizeF, QStandardPaths, QTimer
from PyQt5.QtGui import QBrush, QColor, QFont, QImage, QPainter, QPen, QPixmap, QPolygonF, QTransform
from PyQt5.QtWidgets import QApplication, QDoubleSpinBox, QFrame, QGraphicsPixmapItem, QGraphicsScene, QGraphicsView, QGridLayout, QLabel, QLayout, QHBoxLayout, QPushButton, QScrollArea, QSizePolicy, QSlider, QSpinBox, QSplitter, QSplitterHandle, QToolButton, QToolTip, QVBoxLayout, QWidget
PYQT6 = False

import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading


PRETTY_PARAMETER_NAMES = {
  'frameGapComparision': 'Compare frame n with frame n+?',
  'thresForDetectMovementWithRawVideo': 'Threshold applied on frame n+? minus frame n',
  'halfDiameterRoiBoutDetect': 'Size of sub-frame used for the comparison',
  'minNbPixelForDetectMovementWithRawVideo': 'Minimum number of white pixels in frame n+? minus frame n for bout detection',
  'headEmbededAutoSet_BackgroundExtractionOption': 'Background Extraction',
  'overwriteFirstStepValue': 'Minimum number of pixels between subsequent points',
  'overwriteLastStepValue': 'Maximum number of pixels between subsequent points',
  'overwriteHeadEmbededParamGaussianBlur': 'Gaussian blur applied on the image',
}

TITLE_FONT = QFont('Helvetica', 18, QFont.Weight.Bold, True)
LIGHT_YELLOW = '#FFFFE0'
LIGHT_CYAN = '#E0FFFF'
LIGHT_GREEN = '#90ee90'
GOLD = '#FFD700'
DEFAULT_BUTTON_COLOR = LIGHT_GREEN
SPINBOX_STYLESHEET = '''
QSpinBox::down-button {
  subcontrol-origin: border;
  subcontrol-position: center left;
  height: 20;
  width: 20;
}

QSpinBox::up-button {
  subcontrol-origin: border;
  subcontrol-position: center right;
  height: 20;
  width: 20;
}

QDoubleSpinBox::down-button {
  subcontrol-origin: border;
  subcontrol-position: center left;
  height: 20;
  width: 20;
}

QDoubleSpinBox::up-button {
  subcontrol-origin: border;
  subcontrol-position: center right;
  height: 20;
  width: 20;
}'''


def apply_style(widget, **kwargs):
    font = kwargs.pop('font', None)
    if font is not None:
        widget.setFont(font)
    widget.setStyleSheet(';'.join('%s: %s' % (prop.replace('_', '-'), val)  for prop, val in kwargs.items()))
    return widget


def transformCoordinates(fromRect, toRect, point):
  transform = QTransform()
  QTransform.quadToQuad(QPolygonF((fromRect.topLeft(), fromRect.topRight(), fromRect.bottomLeft(), fromRect.bottomRight())),
                        QPolygonF((toRect.topLeft(), toRect.topRight(), toRect.bottomLeft(), toRect.bottomRight())), transform)
  return transform.map(point)


def _dialogClosed(loop, fn):
  def inner(*args, **kwargs):
    loop.exit()
    return fn(*args, **kwargs)
  return inner


def _getButtonsLayout(buttons, loop, dialog=None):
  buttonsLayout = QHBoxLayout()
  buttonsLayout.addStretch()

  def callback(cb):
    if cb is not None:
      cb()
    if dialog is not None:
      dialog.close()
    else:
      loop.exit()

  for text, *args in buttons:
    enabledSignal = None
    color = None
    if len(args) == 1:
      cb, = args
      exitLoop = True
    elif len(args) == 2:
      cb, exitLoop = args
    elif len(args) == 3:
      cb, exitLoop, enabledSignal = args
    else:
      assert len(args) == 4
      cb, exitLoop, enabledSignal, color = args
    button = QPushButton(text) if color is None else apply_style(QPushButton(text), background_color=color)
    if enabledSignal is not None:
      button.setEnabled(False)
      enabledSignal.connect(lambda enabled, btn=button: btn.setEnabled(bool(enabled)))
    if exitLoop:
      button.clicked.connect(lambda *args, cb=cb: callback(cb))
    else:
      button.clicked.connect(cb)
    buttonsLayout.addWidget(button, alignment=Qt.AlignmentFlag.AlignCenter)
  buttonsLayout.addStretch()
  return buttonsLayout


def showBlockingPage(layout, title=None, buttons=(), dialog=False, labelInfo=None, exitSignals=()):
  loop = QEventLoop()
  for signal in exitSignals:
    signal.connect(lambda *args: loop.exit())
  mainLayout = QVBoxLayout()
  if title is not None:
    titleLabel = apply_style(QLabel(title), font=TITLE_FONT)
    titleLabel.setMinimumSize(1, 1)
    titleLabel.resizeEvent = lambda evt: titleLabel.setMinimumWidth(evt.size().width()) or titleLabel.setWordWrap(evt.size().width() <= titleLabel.sizeHint().width())
    mainLayout.addWidget(titleLabel, alignment=Qt.AlignmentFlag.AlignCenter)
  mainLayout.addLayout(layout)
  mainLayout.addLayout(_getButtonsLayout(buttons, loop))
  app = QApplication.instance()
  assert app is not None
  temporaryPage = QWidget()
  temporaryPage.setLayout(mainLayout)
  stackedLayout = app.window.centralWidget().layout()
  stackedLayout.addWidget(temporaryPage)
  oldWidget = stackedLayout.currentWidget()
  with app.suppressBusyCursor():
    stackedLayout.setCurrentWidget(temporaryPage)
    if labelInfo is not None:
      if len(labelInfo) == 2:
        img, label = labelInfo
        zoomable = False
      else:
        assert len(labelInfo) == 3
        img, label, zoomable = labelInfo
      label.setMinimumSize(1, 1)
      label.show()
      setPixmapFromCv(img, label, zoomable=zoomable)
    loop.exec()
    stackedLayout.setCurrentWidget(oldWidget)
  stackedLayout.removeWidget(temporaryPage)


def showInProgressPage(text):
  def decorator(fn):
    def inner(*args, **kwargs):
      app = QApplication.instance()
      layout = QVBoxLayout()
      layout.addWidget(apply_style(QLabel("%s in progress..." % text), font=TITLE_FONT), alignment=Qt.AlignmentFlag.AlignCenter)
      temporaryPage = QWidget()
      temporaryPage.setLayout(layout)
      stackedLayout = app.window.centralWidget().layout()
      oldWidget = stackedLayout.currentWidget()
      stackedLayout.addWidget(temporaryPage)
      stackedLayout.setCurrentWidget(temporaryPage)
      QAbstractEventDispatcher.instance().processEvents(QEventLoop.ProcessEventsFlag.AllEvents)  # XXX: this is ugly, but required to make sure the page is displayed before the function is executed, since the GUI will get blocked while the function is executing
      try:
        retval = fn(*args, **kwargs)  # XXX: GUI and non-GUI parts are still mixed so this function is expected to set the new page
      finally:
        stackedLayout.removeWidget(temporaryPage)
      return retval
    return inner
  return decorator


def showDialog(layout, title=None, buttons=(), labelInfo=None, timeout=None, exitSignals=()):
  dialog = QWidget()
  for signal in exitSignals:
    signal.connect(lambda *args: dialog.close())
  loop = QEventLoop()
  mainLayout = QVBoxLayout()
  mainLayout.addLayout(layout)
  mainLayout.addLayout(_getButtonsLayout(buttons, loop, dialog=dialog))
  app = QApplication.instance()
  if hasattr(app, 'registerWindow'):
    app.registerWindow(dialog)
  dialog.setWindowTitle(title)
  dialog.move(0, 0)
  dialog.setWindowModality(Qt.WindowModality.ApplicationModal)
  dialog.setLayout(mainLayout)
  if labelInfo is not None:
    if len(labelInfo) == 2:
      img, label = labelInfo
      zoomable = False
    else:
      assert len(labelInfo) == 3
      img, label, zoomable = labelInfo
    height, width = img.shape[:2]
    label.setMinimumSize(width, height)
    layoutSize = mainLayout.totalSizeHint()
    label.setMinimumSize(1, 1)
  else:
    layoutSize = mainLayout.totalSizeHint()
  screenSize = QApplication.primaryScreen().availableSize()
  if hasattr(app, 'window'):
    screenSize -= app.window.frameSize() - app.window.size()
  if layoutSize.width() > screenSize.width() or layoutSize.height() > screenSize.height():
    layoutSize.scale(screenSize, Qt.AspectRatioMode.KeepAspectRatio)
  dialog.setFixedSize(layoutSize)
  dialog.show()
  if labelInfo is not None:
    if len(labelInfo) == 2:
      img, label = labelInfo
      zoomable = False
    else:
      assert len(labelInfo) == 3
      img, label, zoomable = labelInfo
    setPixmapFromCv(img, label, preferredSize=QSize(width, height).scaled(label.size(), Qt.AspectRatioMode.KeepAspectRatio), zoomable=zoomable)
  dialog.closeEvent = _dialogClosed(loop, dialog.closeEvent)
  if timeout is not None:
    QTimer.singleShot(timeout, lambda: dialog.close())
  loop.exec()

def _cvToPixmap(img):
  if len(img.shape) == 2:
    height, width = img.shape
    fmt = QImage.Format.Format_Grayscale8
    bytesPerLine = width
  else:
    assert len(img.shape) == 3
    height, width, channels = img.shape
    fmt = QImage.Format.Format_BGR888
    bytesPerLine = width * channels
  return QPixmap.fromImage(QImage(img.data.tobytes(), width, height, bytesPerLine, fmt))


class _ZoomableImage(QGraphicsView):
  def __init__(self, parent=None):
    super().__init__(parent)
    self._zoom = 0
    self._scene = QGraphicsScene(self)
    self._pixmap = QGraphicsPixmapItem()
    self._scene.addItem(self._pixmap)
    self.setScene(self._scene)
    self.setTransformationAnchor(QGraphicsView.ViewportAnchor.AnchorUnderMouse)
    self.setResizeAnchor(QGraphicsView.ViewportAnchor.AnchorUnderMouse)
    self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
    self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
    self.setFrameShape(QFrame.Shape.NoFrame)
    self._dragging = False
    self._tooltipShown = False

  def fitInView(self):
    rect = QRectF(self._pixmap.pixmap().rect())
    self.setSceneRect(rect)
    unity = self.transform().mapRect(QRectF(0, 0, 1, 1))
    self.scale(1 / unity.width(), 1 / unity.height())
    viewrect = self.viewport().rect()
    scenerect = self.transform().mapRect(rect)
    factor = min(viewrect.width() / scenerect.width(),
                 viewrect.height() / scenerect.height())
    self.scale(factor, factor)
    self._zoom = 0

  def setPixmap(self, pixmap):
    self._zoom = 0
    self.setDragMode(QGraphicsView.DragMode.ScrollHandDrag)
    self._pixmap.setPixmap(pixmap)
    self.fitInView()

  def _update(self, scaleFactor):
    if self._zoom > 0:
      self.scale(scaleFactor, scaleFactor)
    elif self._zoom == 0:
      self.fitInView()
    else:
      self._zoom = 0

  def wheelEvent(self, evt):
    if evt.angleDelta().y() > 0:
      self._zoom += 1
      self._update(1.25)
    else:
      self._zoom -= 1
      self._update(0.8)

  def keyPressEvent(self, evt):
    if evt.modifiers() & Qt.KeyboardModifier.ControlModifier:
      if evt.key() == Qt.Key.Key_Plus:
        self._zoom += 1
        self._update(1.25)
        return
      if evt.key() == Qt.Key.Key_Minus:
        self._zoom -= 1
        self._update(0.8)
        return
    super().keyPressEvent(evt)

  def mouseMoveEvent(self, evt):
    if evt.buttons() == Qt.MouseButton.LeftButton and not self._dragging:
      self._dragging = True
      QApplication.setOverrideCursor(Qt.CursorShape.ClosedHandCursor)
    super().mouseMoveEvent(evt)

  def enterEvent(self, evt):
    QApplication.setOverrideCursor(Qt.CursorShape.CrossCursor)
    super().enterEvent(evt)

  def leaveEvent(self, evt):
    QApplication.restoreOverrideCursor()
    super().leaveEvent(evt)


class _ZoomableImagePoint(_ZoomableImage):
  pointSelected = pyqtSignal(QPoint)
  proceed = pyqtSignal()

  def __init__(self, parent=None):
    super().__init__(parent)
    self._point = None

  def keyPressEvent(self, evt):
    if self._point is not None and (evt.key() == Qt.Key.Key_Enter or evt.key() == Qt.Key.Key_Return):
      self.proceed.emit()
      return
    super().keyPressEvent(evt)

  def mouseReleaseEvent(self, evt):
    if not self._dragging:
      if self._pixmap.isUnderMouse():
        self._point = self.mapToScene(evt.pos()).toPoint()
        self.viewport().update()
        self.pointSelected.emit(self._point)
        if not self._tooltipShown and self._point is not None:
          QToolTip.showText(self.mapToGlobal(self._point), "If you aren't satisfied with the selection, click again.", self, self.rect(), 5000)
          self._tooltipShown = True
    else:
      self._dragging = False
      QApplication.restoreOverrideCursor()
    super().mouseReleaseEvent(evt)

  def paintEvent(self, evt):
    super().paintEvent(evt)
    if self._point is None:
      return
    qp = QPainter(self.viewport())
    if self._point is not None:
      qp.setBrush(QColor(255, 0, 0))
      qp.setPen(Qt.PenStyle.NoPen)
      qp.drawEllipse(self.mapFromScene(QPointF(self._point)), 2, 2)
    qp.end()


class _ZoomableImageCircle(_ZoomableImage):
  circleSelected = pyqtSignal(bool)

  def __init__(self):
    super().__init__()
    self._center = None
    self._currentPosition = None
    self._borderPoint = None
    self.viewport().setMouseTracking(True)

  def mouseReleaseEvent(self, evt):
    if not self._dragging:
      if self._pixmap.isUnderMouse():
        if self._center is None or self._borderPoint is not None:
          self._center = self.mapToScene(evt.pos()).toPoint()
          self._borderPoint = None
          self._currentPosition = None
          self.circleSelected.emit(False)
        else:
          self._borderPoint = self.mapToScene(evt.pos()).toPoint()
          self.circleSelected.emit(True)
        self.viewport().update()
        if not self._tooltipShown and self._center is not None and self._borderPoint is not None:
          QToolTip.showText(self.mapToGlobal(self._center), "If you aren't satisfied with the selection, click again.", self, self.rect(), 5000)
          self._tooltipShown = True
    else:
      self._dragging = False
      QApplication.restoreOverrideCursor()
    super().mouseReleaseEvent(evt)

  def mouseMoveEvent(self, evt):
    super().mouseMoveEvent(evt)
    if not self._dragging:
      self._currentPosition = self.mapToScene(evt.pos()).toPoint()
      self.viewport().update()

  def leaveEvent(self, evt):
    self._currentPosition = None
    self.viewport().update()
    super().leaveEvent(evt)

  def paintEvent(self, evt):
    super().paintEvent(evt)
    if self._currentPosition is None and self._center is None:
      return
    qp = QPainter(self.viewport())
    if self._center is not None:
      qp.setBrush(QColor(255, 0, 0))
      qp.setPen(Qt.PenStyle.NoPen)
      borderPoint = self._borderPoint if self._borderPoint is not None else self._currentPosition if self._currentPosition is not None else None
      mappedCenter = self.mapFromScene(self._center)
      qp.drawEllipse(mappedCenter, 2, 2)
      if borderPoint is not None:
        delta = self.mapFromScene(borderPoint) - mappedCenter
        radius = int(math.sqrt(delta.x() * delta.x() + delta.y() * delta.y()))
        qp.setBrush(Qt.BrushStyle.NoBrush)
        qp.setPen(QColor(0, 0, 255))
        qp.drawEllipse(mappedCenter, radius, radius)
    qp.end()


def setPixmapFromCv(img, label, preferredSize=None, zoomable=False):
  originalPixmap = _cvToPixmap(img)
  if not label.isVisible():
    label.setPixmap(originalPixmap)
    return
  scaling = label.devicePixelRatio() if PYQT6 else label.devicePixelRatioF()
  if label.pixmap() is None or label.pixmap().isNull():
    label.hide()
    label.setPixmap(originalPixmap)
    label.show()
  if preferredSize is None:
    preferredSize = label.pixmap().size()
  size = preferredSize.scaled(label.size(), Qt.AspectRatioMode.KeepAspectRatio)
  if not zoomable:
    img = cv2.resize(img, (int(size.width() * scaling), int(size.height() * scaling)))
    pixmap = _cvToPixmap(img)
    pixmap.setDevicePixelRatio(scaling)
    label.setPixmap(pixmap)
  else:
    label.hide()
    image = _ZoomableImagePoint() if not isinstance(label, _InteractiveLabelCircle) else _ZoomableImageCircle()
    image.sizeHint = lambda: size
    image.setMaximumSize(size)
    image.viewport().setFixedSize(size)
    image.setPixmap(originalPixmap)
    label.parentWidget().layout().replaceWidget(label, image)
    image.setFocus()
    if hasattr(image, "pointSelected"):
      def pointSelected(point):
        label.pointSelected.emit(point)
        label.getCoordinates = lambda: (point.x(), point.y())
      image.pointSelected.connect(pointSelected)
      image.proceed.connect(label.proceed.emit)
    if hasattr(image, "circleSelected"):
      def circleSelected(selected):
        label.circleSelected.emit(selected)
      image.circleSelected.connect(circleSelected)
      def calculateRadius():
        if image._borderPoint is None or image._center is None:
          return None, None
        delta = image._borderPoint - image._center
        return image._center, int(math.sqrt(delta.x() * delta.x() + delta.y() * delta.y()))
      label.getInfo = calculateRadius


class _DoubleSlider(QSlider):
  _DECIMALS = 2

  def __init__(self, *args, **kargs):
    super().__init__( *args, **kargs)
    self._factor = 10 ** self._DECIMALS

  def value(self):
    return super().value() / self._factor

  def setMinimum(self, value):
    return super().setMinimum(int(value * self._factor))

  def setMaximum(self, value):
    return super().setMaximum(int(value * self._factor))

  def setRange(self, minimum, maximum):
    return super().setRange(int(minimum * self._factor), int(maximum * self._factor))

  def setValue(self, value):
    super().setValue(int(value * self._factor))


class SliderWithSpinbox(QWidget):
  valueChanged = pyqtSignal(int)

  def __init__(self, value, minimum, maximum, name=None, double=False):
    super().__init__()
    minimum = int(minimum)
    maximum = int(maximum)

    layout = QGridLayout()
    layout.setRowStretch(0, 1)
    layout.setColumnStretch(0, 1)
    layout.setRowStretch(3, 1)
    layout.setColumnStretch(5, 1)
    layout.setVerticalSpacing(0)

    minLabel = QLabel(str(minimum))
    layout.addWidget(minLabel, 1, 1, Qt.AlignmentFlag.AlignLeft)
    maxLabel = QLabel(str(maximum))
    layout.addWidget(maxLabel, 1, 3, Qt.AlignmentFlag.AlignRight)
    slider = QSlider(Qt.Orientation.Horizontal) if not double else _DoubleSlider(Qt.Orientation.Horizontal)
    slider.setMinimumWidth(350)
    slider.setRange(minimum, maximum)
    slider.setValue(value)
    layout.addWidget(slider, 2, 1, 1, 3)

    spinbox = QSpinBox() if not double else QDoubleSpinBox()
    if double:
      spinbox.setSingleStep(0.1)
    spinbox.setStyleSheet(SPINBOX_STYLESHEET)
    spinbox.setMinimumWidth(90)
    spinbox.setRange(minimum, maximum)
    spinbox.setValue(value)
    layout.addWidget(spinbox, 2, 4)

    def spinboxValueChanged():
      value = spinbox.value()
      blocked = slider.blockSignals(True)
      slider.setValue(value)
      slider.blockSignals(blocked)
      self.valueChanged.emit(value)
    spinbox.valueChanged.connect(spinboxValueChanged)
    slider.valueChanged.connect(lambda: spinbox.setValue(slider.value()))

    layout.setContentsMargins(20, 5, 20, 5)
    self.setLayout(layout)

    if name is not None:
      self.setFixedWidth(layout.totalSizeHint().width())
      titleLabel = QLabel(PRETTY_PARAMETER_NAMES.get(name, name))
      titleLabel.setMinimumSize(1, 1)
      titleLabel.resizeEvent = lambda evt: titleLabel.setMinimumWidth(evt.size().width()) or titleLabel.setWordWrap(evt.size().width() <= titleLabel.sizeHint().width())
      slider.rangeChanged.connect(lambda: titleLabel.setMinimumWidth(1) or titleLabel.setWordWrap(False))
      layout.addWidget(titleLabel, 1, 2, Qt.AlignmentFlag.AlignCenter)

    self.value = spinbox.value
    self.minimum = spinbox.minimum
    self.maximum = spinbox.maximum
    self.setValue = lambda value: spinbox.setValue(value) or slider.setValue(value)
    self.setMinimum = lambda min_: spinbox.setMinimum(min_) or slider.setMinimum(min_) or minLabel.setText(str(int(min_)))
    self.setMaximum = lambda max_: spinbox.setMaximum(max_) or slider.setMaximum(max_) or maxLabel.setText(str(int(max_)))
    self.setRange = lambda min_, max_: spinbox.setRange(min_, max_) or slider.setRange(min_, max_)
    self.setSingleStep = lambda step: spinbox.setSingleStep(step) or slider.setSingleStep(step)
    self.isSliderDown = slider.isSliderDown
    self.sliderWidth = slider.width
    self.setPosition = slider.setSliderPosition


def _chooseFrameLayout(cap, spinboxValues, title, titleStyle=None):
  if titleStyle is None:
    titleStyle = {'font': TITLE_FONT}
  layout = QVBoxLayout()
  titleLabel = apply_style(QLabel(title), **titleStyle)
  titleLabel.setMinimumSize(1, 1)
  titleLabel.resizeEvent = lambda evt: titleLabel.setMinimumWidth(evt.size().width()) or titleLabel.setWordWrap(evt.size().width() <= titleLabel.sizeHint().width())
  layout.addWidget(titleLabel, alignment=Qt.AlignmentFlag.AlignCenter)
  video = QLabel()
  layout.addWidget(video, alignment=Qt.AlignmentFlag.AlignCenter, stretch=1)

  firstFrame, minFrame, maxFrame = spinboxValues
  frameSlider = SliderWithSpinbox(firstFrame, minFrame, maxFrame, name="Frame")

  def getFrame():
    cap.set(1, frameSlider.value())
    ret, img = cap.read()
    return img
  frameSlider.valueChanged.connect(lambda: setPixmapFromCv(getFrame(), video))

  sublayout = QHBoxLayout()
  sublayout.addStretch(1)
  sublayout.addWidget(frameSlider, alignment=Qt.AlignmentFlag.AlignCenter)
  if maxFrame > 1000:
    adjustLayout = QVBoxLayout()
    adjustLayout.setSpacing(0)
    adjustLayout.addStretch()
    zoomInSliderBtn = QPushButton("Zoom in slider")

    def updatePreciseFrameSlider(value):
      if frameSlider.minimum() == value and frameSlider.minimum():
        frameSlider.setMinimum(frameSlider.minimum() - 1)
        frameSlider.setMaximum(frameSlider.maximum() - 1)
      elif value == frameSlider.maximum() and frameSlider.maximum() != maxFrame:
        frameSlider.setMinimum(frameSlider.minimum() + 1)
        frameSlider.setMaximum(frameSlider.maximum() + 1)

    def zoomInButtonClicked():
      if "in" in zoomInSliderBtn.text():
        zoomInSliderBtn.setText("Zoom out slider")
        value = frameSlider.value()
        minimum = value - 250
        maximum = value + 250
        if minimum < 0:
          maximum = 500
          minimum = 0
        if maximum > frameSlider.maximum():
          maximum = frameSlider.maximum()
          minimum = maximum - 500
        frameSlider.setMinimum(max(0, minimum))
        frameSlider.setMaximum(min(frameSlider.maximum(), maximum))
        frameSlider.setValue(value)
        frameSlider.valueChanged.connect(updatePreciseFrameSlider)
      else:
        zoomInSliderBtn.setText("Zoom in slider")
        frameSlider.setMinimum(0)
        frameSlider.setMaximum(maxFrame)
        frameSlider.valueChanged.disconnect(updatePreciseFrameSlider)
    zoomInSliderBtn.clicked.connect(zoomInButtonClicked)
    adjustLayout.addWidget(QLabel())
    adjustLayout.addWidget(zoomInSliderBtn, alignment=Qt.AlignmentFlag.AlignLeft, stretch=1)
    adjustLayout.addStretch()
    sublayout.addLayout(adjustLayout, stretch=1)
  else:
    sublayout.addStretch(1)
  layout.addLayout(sublayout)

  return layout, video, frameSlider

def chooseBeginningPage(app, videoPath, title, chooseFrameBtnText, chooseFrameBtnCb, extraButtonInfo=None, titleStyle=None, additionalLayout=None):
  cap = zzVideoReading.VideoCapture(videoPath)
  cap.set(1, 1)
  ret, frame = cap.read()
  layout, label, valueWidget = _chooseFrameLayout(cap, (1, 0, cap.get(7) - 2), title, titleStyle=titleStyle)
  if additionalLayout is not None:
    layout.addLayout(additionalLayout)

  buttonsLayout = QHBoxLayout()
  buttonsLayout.addStretch()
  if app.configFileHistory:
    backBtn = QPushButton("Back")
    backBtn.setObjectName("back")
    buttonsLayout.addWidget(backBtn)
  chooseFrameBtn = QPushButton(chooseFrameBtnText) if extraButtonInfo is not None else apply_style(QPushButton(chooseFrameBtnText), background_color=DEFAULT_BUTTON_COLOR)
  def chooseFrameBtnClicked():
    app.configFile["firstFrame"] = valueWidget.value()
    app.configFile["firstFrameForBackExtract"] = valueWidget.value()
    chooseFrameBtnCb()
  chooseFrameBtn.clicked.connect(chooseFrameBtnClicked)
  buttonsLayout.addWidget(chooseFrameBtn)
  extraBtn = None
  if extraButtonInfo is not None:
    if len(extraButtonInfo) == 2:
      text, cb = extraButtonInfo
      styleKwargs = {}
    else:
      assert len(extraButtonInfo) == 3
      text, cb, styleKwargs = extraButtonInfo
    extraBtn = apply_style(QPushButton(text), **styleKwargs)
    extraBtn.clicked.connect(cb)
    buttonsLayout.addWidget(extraBtn)
  buttonsLayout.addStretch()
  layout.addLayout(buttonsLayout)
  page = QWidget()
  page.setLayout(layout)
  stackedLayout = app.window.centralWidget().layout()
  stackedLayout.addWidget(page)
  oldWidget = stackedLayout.currentWidget()
  with app.suppressBusyCursor():
    stackedLayout.setCurrentWidget(page)
    label.setMinimumSize(1, 1)
    label.show()
    setPixmapFromCv(frame, label)
  buttons = []
  if app.configFileHistory:
    buttons.append(backBtn)
  buttons.append(chooseFrameBtn)
  if extraBtn is not None:
    buttons.append(extraBtn)
  for btn in buttons:
    btn.clicked.connect(lambda: stackedLayout.removeWidget(page))


def chooseEndPage(app, videoPath, title, chooseFrameBtnText, chooseFrameBtnCb):
  cap = zzVideoReading.VideoCapture(videoPath)
  maximum = cap.get(7) - 2
  cap.set(1, maximum)
  ret, frame = cap.read()
  layout, label, valueWidget = _chooseFrameLayout(cap, (maximum, app.configFile["firstFrame"] + 1, maximum), title)

  buttonsLayout = QHBoxLayout()
  buttonsLayout.addStretch()
  if app.configFileHistory:
    backBtn = QPushButton("Back")
    backBtn.setObjectName("back")
    buttonsLayout.addWidget(backBtn)
  chooseFrameBtn = apply_style(QPushButton(chooseFrameBtnText), background_color=DEFAULT_BUTTON_COLOR)
  def chooseFrameBtnClicked():
    app.configFile["lastFrame"] = valueWidget.value()
    app.configFile["lastFrameForBackExtract"] = valueWidget.value()
    chooseFrameBtnCb()
  chooseFrameBtn.clicked.connect(chooseFrameBtnClicked)
  buttonsLayout.addWidget(chooseFrameBtn)
  buttonsLayout.addStretch()
  layout.addLayout(buttonsLayout)
  page = QWidget()
  page.setLayout(layout)
  stackedLayout = app.window.centralWidget().layout()
  stackedLayout.addWidget(page)
  oldWidget = stackedLayout.currentWidget()
  with app.suppressBusyCursor():
    stackedLayout.setCurrentWidget(page)
    label.setMinimumSize(1, 1)
    label.show()
    setPixmapFromCv(frame, label)
  for btn in (backBtn, chooseFrameBtn) if app.configFileHistory else (chooseFrameBtn,):
    btn.clicked.connect(lambda: stackedLayout.removeWidget(page))


class _InteractiveLabelPoint(QLabel):
  pointSelected = pyqtSignal(QPoint)
  proceed = pyqtSignal()

  def __init__(self, width, height, selectingRegion):
    super().__init__()
    self._width = width
    self._height = height
    self._point = None
    self._selectingRegion = selectingRegion
    self._currentPosition = None
    self._tooltipShown = False
    self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
    if self._selectingRegion:
      self.setMouseTracking(True)

  def keyPressEvent(self, evt):
    if self._point is not None and (evt.key() == Qt.Key.Key_Enter or evt.key() == Qt.Key.Key_Return):
      self.proceed.emit()
      return
    super().keyPressEvent(evt)

  def mouseMoveEvent(self, evt):
    if self._selectingRegion:
      self._currentPosition = evt.pos()
      self.update()

  def mousePressEvent(self, evt):
    self._point = evt.pos()
    self.update()
    self.pointSelected.emit(self._point)

  def mouseReleaseEvent(self, evt):
    if self._point is not None and not self._tooltipShown:
      QToolTip.showText(self.mapToGlobal(self._point), "If you aren't satisfied with the selection, click again.", self, self.rect(), 5000)
      self._tooltipShown = True

  def enterEvent(self, evt):
    QApplication.setOverrideCursor(Qt.CursorShape.CrossCursor)

  def leaveEvent(self, evt):
    QApplication.restoreOverrideCursor()
    self._currentPosition = None
    self.update()

  def paintEvent(self, evt):
    super().paintEvent(evt)
    if self._currentPosition is None and self._point is None:
      return
    qp = QPainter()
    qp.begin(self)
    if self._currentPosition is not None:
      qp.setPen(QColor(255, 0, 0))
      qp.drawLine(0, self._currentPosition.y(), self.width(), self._currentPosition.y())
      qp.drawLine(self._currentPosition.x(), 0, self._currentPosition.x(), self.height())
    if self._point is not None:
      qp.setBrush(QColor(255, 0, 0))
      qp.setPen(Qt.PenStyle.NoPen)
      qp.drawEllipse(self._point, 2, 2)
    qp.end()

  def resizeEvent(self, evt):
    super().resizeEvent(evt)
    self._size = self.size()

  def getCoordinates(self):
    if self._point is None:
      return 0, 0
    point = self._point
    if self._size.height() != self._height or self._size.width() != self._width:
      point = transformCoordinates(QRectF(QPointF(0, 0), QSizeF(self._size)), QRectF(QPointF(0, 0), QSizeF(self._width, self._height)), self._point)
    return point.x(), point.y()


def getPoint(frame, title, extraButtons=(), selectingRegion=False, backBtnCb=None, zoomable=False, useNext=True, dialog=False):
  height, width = frame.shape[:2]

  layout = QVBoxLayout()
  additionalText = "Enter/Return keys can be used instead of clicking Next."
  if zoomable:
    controlModifier = 'Ctrl' if sys.platform != 'darwin' else 'Cmd'  # Qt documentation explicilty states this
    additionalText += "\nYou can zoom in/out using the mouse wheel or %s and +/- and drag the image." % controlModifier
  layout.addWidget(QLabel(additionalText), alignment=Qt.AlignmentFlag.AlignCenter)

  video = _InteractiveLabelPoint(width, height, selectingRegion)

  def callback(cb):
    if zoomable:
      zoomableImage = layout.itemAt(1).widget()
      layout.replaceWidget(zoomableImage, video)
      zoomableImage.hide()
      video.show()
    cb(video)

  def updateButton(args):
    text, cb, exitLoop = args[:3]
    retval = (text, lambda: callback(cb), exitLoop)
    if len(args) == 4:
      retval += (video.pointSelected,)
    return retval

  extraButtons = tuple(updateButton(button) for button in extraButtons)
  buttons = (("Back", backBtnCb, True),) if backBtnCb is not None else ()
  buttons += (("Next", None, True, video.pointSelected, DEFAULT_BUTTON_COLOR),) if useNext else ()
  layout.addWidget(video, alignment=Qt.AlignmentFlag.AlignCenter, stretch=1)
  if not useNext:
    video.pointSelected.connect(lambda: QApplication.restoreOverrideCursor())
  if not dialog:
    showBlockingPage(layout, title=title, buttons=extraButtons + buttons, labelInfo=(frame, video, zoomable), exitSignals=(video.proceed,) if useNext else (video.pointSelected,))
  else:
    showDialog(layout, title=title, buttons=extraButtons + buttons, labelInfo=(frame, video, zoomable), exitSignals=(video.proceed,) if useNext else (video.pointSelected,))
  return video.getCoordinates()


class _InteractiveLabelRect(QLabel):
  regionSelected = pyqtSignal(bool)

  def __init__(self, width, height, initialRect):
    super().__init__()
    self._width = width
    self._height = height
    self._currentPosition = None
    if initialRect is not None:
      self._initialRect = initialRect
    self._topLeft = None
    self._bottomRight = None
    self._size = self.size()
    self._tooltipShown = False
    self.setMouseTracking(True)

  def clearRectangle(self):
    self._topLeft = None
    self._bottomRight = None
    self.update()
    self.regionSelected.emit(False)

  def mousePressEvent(self, evt):
    if self._topLeft is None or self._bottomRight is not None:
      self._topLeft = evt.pos()
      self._bottomRight = None
      self._currentPosition = None
      self.regionSelected.emit(False)
    else:
      self._bottomRight = evt.pos()
      self.regionSelected.emit(True)
    self.update()

  def mouseReleaseEvent(self, evt):
    if self._bottomRight is not None and not self._tooltipShown:
      QToolTip.showText(self.mapToGlobal(self._bottomRight), "If you aren't satisfied with the selection, click again.", self, self.rect(), 5000)
      self._tooltipShown = True

  def mouseMoveEvent(self, evt):
    self._currentPosition = evt.pos()
    self.update()

  def enterEvent(self, evt):
    QApplication.setOverrideCursor(Qt.CursorShape.CrossCursor)

  def leaveEvent(self, evt):
    QApplication.restoreOverrideCursor()
    self._currentPosition = None
    self.update()

  def paintEvent(self, evt):
    super().paintEvent(evt)
    if hasattr(self, '_initialRect'):
      self._topLeft, self._bottomRight = (transformCoordinates(QRectF(QPointF(0, 0), QSizeF(self._width, self._height)), QRectF(QPointF(0, 0), QSizeF(self.size())), point) for point in self._initialRect)
      del self._initialRect
    if self._currentPosition is None and self._topLeft is None:
      return
    qp = QPainter()
    qp.begin(self)
    if self._currentPosition is not None and \
        (self._topLeft is None and self._bottomRight is None or self._bottomRight is not None):
      qp.setPen(QColor(255, 0, 0))
      qp.drawLine(0, self._currentPosition.y(), self.width(), self._currentPosition.y())
      qp.drawLine(self._currentPosition.x(), 0, self._currentPosition.x(), self.height())
    if self._topLeft is not None:
      qp.setPen(QColor(0, 0, 255))
      x = self._topLeft.x()
      y = self._topLeft.y()
      bottomRight = self._bottomRight or self._currentPosition
      if bottomRight is None:
        qp.drawPoint(x, y)
      else:
        width = bottomRight.x() - x
        height = bottomRight.y() - y
        qp.drawRect(x, y, width, height)
    qp.end()

  def resizeEvent(self, evt):
    super().resizeEvent(evt)
    if self._topLeft is not None:
      self._topLeft = transformCoordinates(QRectF(QPointF(0, 0), QSizeF(self._size)), QRectF(QPointF(0, 0), QSizeF(evt.size())), self._topLeft)
    if self._bottomRight is not None:
      self._bottomRight = transformCoordinates(QRectF(QPointF(0, 0), QSizeF(self._size)), QRectF(QPointF(0, 0), QSizeF(evt.size())), self._bottomRight)
    if self._currentPosition:
      self._currentPosition = transformCoordinates(QRectF(QPointF(0, 0), QSizeF(self._size)), QRectF(QPointF(0, 0), QSizeF(evt.size())), self._currentPosition)
    self._size = self.size()

  def getCoordinates(self):
    if self._topLeft is None or self._bottomRight is None:
      return [0, 0], [0, 0]
    points = (self._topLeft, self._bottomRight)
    if self._size.height() != self._height or self._size.width() != self._width:
      points = (transformCoordinates(QRectF(QPointF(0, 0), QSizeF(self._size)), QRectF(QPointF(0, 0), QSizeF(self._width, self._height)), point) for point in points)
    return ([point.x(), point.y()] for point in points)


def getRectangle(frame, title, backBtnCb=None, dialog=False, buttons=None, initialRect=None, allowEmpty=False, contrastCheckbox=None, getFrame=None):
  height, width, _ = frame.shape

  layout = QVBoxLayout()

  video = _InteractiveLabelRect(width, height, initialRect)
  layout.addWidget(video, alignment=Qt.AlignmentFlag.AlignCenter, stretch=1)
  if contrastCheckbox is not None:
    assert getFrame is not None
    contrastCheckbox.toggled.connect(lambda checked: setPixmapFromCv(getFrame(checked), video))
    layout.addWidget(contrastCheckbox, alignment=Qt.AlignmentFlag.AlignCenter)
  if allowEmpty:
    clearRectangleBtn = QPushButton('Clear rectangle')
    clearRectangleBtn.clicked.connect(video.clearRectangle)
    clearRectangleBtn.setEnabled(initialRect is not None)
    layout.addWidget(clearRectangleBtn, alignment=Qt.AlignmentFlag.AlignCenter)
    video.regionSelected.connect(clearRectangleBtn.setEnabled)
  if buttons is None:
    if backBtnCb is not None:
      buttons = (("Back", backBtnCb, True), ("Next", None, True, video.regionSelected, DEFAULT_BUTTON_COLOR))
    else:
      buttons = (("Next", None, True, video.regionSelected, DEFAULT_BUTTON_COLOR),)
  else:
    assert backBtnCb is None
  if not dialog:
    showBlockingPage(layout, title=title, buttons=buttons, labelInfo=(frame, video))
  else:
    showDialog(layout, title=title, buttons=buttons, labelInfo=(frame, video))

  return video.getCoordinates()


class _CVLabelLine(QLabel):
  pointSelected = pyqtSignal(QPoint)
  proceed = pyqtSignal()

  def __init__(self, width, height, frame, thickness, startPoint):
    super().__init__()
    self._width = width
    self._height = height
    self._point = None
    self._currentPosition = None
    self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
    self.setMouseTracking(True)
    self._thickness = thickness
    self._startPoint = startPoint
    self._frame = frame

  def keyPressEvent(self, evt):
    if self._point is not None and (evt.key() == Qt.Key.Key_Enter or evt.key() == Qt.Key.Key_Return):
      self.proceed.emit()
      return
    super().keyPressEvent(evt)

  def mouseMoveEvent(self, evt):
    self._currentPosition = evt.pos()
    point = self._currentPosition
    if self._size.height() != self._height or self._size.width() != self._width:
      point = transformCoordinates(QRectF(QPointF(0, 0), QSizeF(self._size)), QRectF(QPointF(0, 0), QSizeF(self._width, self._height)), point)
    setPixmapFromCv(cv2.line(self._frame.copy(), (self._startPoint), (point.x(), point.y()), (255, 255, 255), self._thickness), self)

  def mousePressEvent(self, evt):
    self._point = evt.pos()
    self.pointSelected.emit(self._point)

  def enterEvent(self, evt):
    QApplication.setOverrideCursor(Qt.CursorShape.CrossCursor)

  def leaveEvent(self, evt):
    QApplication.restoreOverrideCursor()
    self._currentPosition = None

  def resizeEvent(self, evt):
    super().resizeEvent(evt)
    self._size = self.size()

  def getCoordinates(self):
    if self._point is None:
      return 0, 0
    point = self._point
    if self._size.height() != self._height or self._size.width() != self._width:
      point = transformCoordinates(QRectF(QPointF(0, 0), QSizeF(self._size)), QRectF(QPointF(0, 0), QSizeF(self._width, self._height)), self._point)
    return point.x(), point.y()


def getLine(frame, title, thickness, startPoint):
  height, width = frame.shape[:2]

  layout = QVBoxLayout()

  video = _CVLabelLine(width, height, frame, thickness, startPoint)
  layout.addWidget(video, alignment=Qt.AlignmentFlag.AlignCenter)
  buttons = (("Next", None, True),)
  showDialog(layout, title=title, buttons=buttons, labelInfo=(frame, video), exitSignals=(video.pointSelected,))

  return video.getCoordinates()


class _CVDrawLabel(QLabel):
  proceed = pyqtSignal()

  def __init__(self, width, height, frame):
    super().__init__()
    self._width = width
    self._height = height
    self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
    self._frame = frame

  def keyPressEvent(self, evt):
    if self._point is not None and (evt.key() == Qt.Key.Key_Enter or evt.key() == Qt.Key.Key_Return):
      self.proceed.emit()
      return
    super().keyPressEvent(evt)

  def mouseMoveEvent(self, evt):
    point = evt.pos()
    if self._size.height() != self._height or self._size.width() != self._width:
      point = transformCoordinates(QRectF(QPointF(0, 0), QSizeF(self._size)), QRectF(QPointF(0, 0), QSizeF(self._width, self._height)), point)
    self._frame = cv2.circle(self._frame, (point.x(), point.y()), 3, (255, 255, 255), -1)
    setPixmapFromCv(self._frame, self)

  def mousePressEvent(self, evt):
    point = evt.pos()
    if self._size.height() != self._height or self._size.width() != self._width:
      point = transformCoordinates(QRectF(QPointF(0, 0), QSizeF(self._size)), QRectF(QPointF(0, 0), QSizeF(self._width, self._height)), point)
    self._frame = cv2.circle(self._frame, (point.x(), point.y()), 3, (255, 255, 255), -1)
    setPixmapFromCv(self._frame, self)

  def enterEvent(self, evt):
    QApplication.setOverrideCursor(Qt.CursorShape.CrossCursor)

  def leaveEvent(self, evt):
    QApplication.restoreOverrideCursor()

  def resizeEvent(self, evt):
    super().resizeEvent(evt)
    self._size = self.size()

  def getFrame(self):
    return self._frame


def drawPoints(frame, title):
  height, width = frame.shape[:2]

  layout = QVBoxLayout()

  video = _CVDrawLabel(width, height, frame)
  layout.addWidget(video, alignment=Qt.AlignmentFlag.AlignCenter)
  buttons = (("Done", None, True),)
  showDialog(layout, title=title, buttons=buttons, labelInfo=(frame, video))
  return video.getFrame()


class _FrameLabel(QLabel):
  proceed = pyqtSignal()

  def keyPressEvent(self, evt):
    super().keyPressEvent(evt)
    self.proceed.emit()


def showFrame(frame, title='', buttons=(), timeout=None):
  label = _FrameLabel()
  label.setMinimumSize(1, 1)
  label.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
  layout = QVBoxLayout()
  layout.addWidget(label, alignment=Qt.AlignmentFlag.AlignCenter)
  showDialog(layout, title=title, buttons=buttons, labelInfo=(frame, label), timeout=timeout, exitSignals=(label.proceed,))


def addToHistory(fn):
  def inner(*args, **kwargs):
    app = QApplication.instance()
    configFileState = app.configFile.copy()
    def restoreState(restoreConfig=True):
      if restoreConfig:
        app.configFile.clear()
        app.configFile.update(configFileState)
      del app.configFileHistory[-1:]
      fn(*args, **kwargs)
    app.configFileHistory.append(restoreState)
    return fn(*args, **kwargs)
  return inner


class Expander(QWidget):
  def __init__(self, parent, title, layout, animationDuration=200, showFrame=False, addScrollbars=False):
    super().__init__()

    self._toggleButton = toggleButton = QToolButton()
    toggleButton.setStyleSheet("QToolButton { border: none; }")
    toggleButton.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
    toggleButton.setArrowType(Qt.ArrowType.RightArrow)
    toggleButton.setText(str(title))
    toggleButton.setCheckable(True)
    toggleButton.setChecked(False)

    self._contentArea = contentArea = QScrollArea()
    if not showFrame:
      contentArea.setFrameShape(QFrame.Shape.NoFrame)
    contentArea.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)
    contentArea.setMaximumHeight(0)
    contentArea.setMinimumHeight(0)

    mainLayout = QGridLayout()
    mainLayout.setVerticalSpacing(0)
    mainLayout.setContentsMargins(0, 0, 0, 0)
    mainLayout.addWidget(toggleButton, 0, 0, 1, 3, Qt.AlignmentFlag.AlignHCenter)
    mainLayout.addWidget(contentArea, 1, 0, 1, 3)
    self.setLayout(mainLayout)

    if not addScrollbars:
      contentArea.setLayout(layout)
    else:
      widget = QWidget(self)
      widget.setLayout(layout)
      contentArea.setWidgetResizable(True)
      contentArea.setWidget(widget)
    self._collapseHeight = self.sizeHint().height() - contentArea.maximumHeight()
    contentHeight = layout.sizeHint().height()
    self._toggleAnimation = toggleAnimation = QParallelAnimationGroup()
    toggleAnimation.addAnimation(QPropertyAnimation(self, b"minimumHeight"))
    toggleAnimation.addAnimation(QPropertyAnimation(self, b"maximumHeight"))
    toggleAnimation.addAnimation(QPropertyAnimation(contentArea, b"maximumHeight"))
    for i in range(toggleAnimation.animationCount() - 1):
      spoilerAnimation = toggleAnimation.animationAt(i)
      spoilerAnimation.setDuration(animationDuration)
      spoilerAnimation.setStartValue(self._collapseHeight)
      spoilerAnimation.setEndValue(self._collapseHeight + contentHeight)
    contentAnimation = toggleAnimation.animationAt(toggleAnimation.animationCount() - 1)
    contentAnimation.setDuration(animationDuration)
    contentAnimation.setStartValue(0)
    contentAnimation.setEndValue(contentHeight)

    def startAnimation(checked):
      arrowType = Qt.ArrowType.DownArrow if checked else Qt.ArrowType.RightArrow
      direction = QAbstractAnimation.Direction.Forward if checked else QAbstractAnimation.Direction.Backward
      toggleButton.setArrowType(arrowType)
      toggleAnimation.setDirection(direction)
      toggleAnimation.start()
    toggleButton.clicked.connect(startAnimation)

  def refresh(self, availableHeight):
    layout = self._contentArea.layout()
    height = layout.sizeHint().height() if layout is not None else self._contentArea.widget().sizeHint().height() + 5
    contentHeight = min(height, availableHeight - self._collapseHeight - 10)
    if self._contentArea.maximumHeight():
      self._contentArea.setMaximumHeight(contentHeight)
      self.setMinimumHeight(self._collapseHeight + contentHeight)
      self.setMaximumHeight(self._collapseHeight + contentHeight)
    for i in range(self._toggleAnimation.animationCount() - 1):
      self._toggleAnimation.animationAt(i).setEndValue(self._collapseHeight + contentHeight)
    self._toggleAnimation.animationAt(self._toggleAnimation.animationCount() - 1).setEndValue(contentHeight)


class _InteractiveLabelCircle(QLabel):
  circleSelected = pyqtSignal(bool)

  def __init__(self, width, height):
    super().__init__()
    self._width = width
    self._height = height
    self._center = None
    self._currentPosition = None
    self._radius = None
    self._size = None
    self._tooltipShown = False
    self.setMouseTracking(True)

  def mousePressEvent(self, evt):
    if self._center is None or self._radius is not None:
      self._center = evt.pos()
      self._radius = None
      self._currentPosition = None
      self.circleSelected.emit(False)
    else:
      self._radius = QLine(self._center, evt.pos())
      self.circleSelected.emit(True)
    self.update()

  def mouseReleaseEvent(self, evt):
    if self._radius is not None and not self._tooltipShown:
      QToolTip.showText(evt.globalPosition() if PYQT6 else evt.globalPos(), "If you aren't satisfied with the selection, click again.", self, self.rect(), 5000)
      self._tooltipShown = True

  def mouseMoveEvent(self, evt):
    self._currentPosition = evt.pos()
    self.update()

  def enterEvent(self, evt):
    QApplication.setOverrideCursor(Qt.CursorShape.CrossCursor)

  def leaveEvent(self, evt):
    QApplication.restoreOverrideCursor()
    self._currentPosition = None
    self.update()

  def paintEvent(self, evt):
    super().paintEvent(evt)
    if self._currentPosition is None and self._center is None:
      return
    qp = QPainter()
    qp.begin(self)
    if self._center is not None:
      qp.setBrush(QColor(255, 0, 0))
      qp.setPen(Qt.PenStyle.NoPen)
      radius = self._radius if self._radius is not None else QLine(self._center, self._currentPosition) if self._currentPosition is not None else None
      qp.drawEllipse(self._center, 2, 2)
      if radius is not None:
        radius = math.sqrt(radius.dx() * radius.dx() + radius.dy() * radius.dy())
        qp.setBrush(Qt.BrushStyle.NoBrush)
        qp.setPen(QColor(0, 0, 255))
        qp.drawEllipse(self._center, radius, radius)
    qp.end()

  def resizeEvent(self, evt):
    super().resizeEvent(evt)
    self._size = self.size()

  def getInfo(self):
    if self._center is None or self._radius is None:
      return None, None
    if self._size.height() != self._height or self._size.width() != self._width:
      center = transformCoordinates(QRectF(QPointF(0, 0), QSizeF(self._size)), QRectF(QPointF(0, 0), QSizeF(self._width, self._height)), self._center)
      radius = transformCoordinates(QRectF(QPointF(0, 0), QSizeF(self._size)), QRectF(QPointF(0, 0), QSizeF(self._width, self._height)), self._radius)
    else:
      center = self._center
      radius = self._radius
    return center, int(math.sqrt(radius.dx() * radius.dx() + radius.dy() * radius.dy()))


def getCircle(frame, title, backBtnCb=None, zoomable=False):
  height, width, _ = frame.shape

  layout = QVBoxLayout()

  video = _InteractiveLabelCircle(width, height)
  layout.addWidget(video, alignment=Qt.AlignmentFlag.AlignCenter, stretch=1)
  if backBtnCb is not None:
    buttons = (("Cancel", backBtnCb, True), ("Ok", None, True, video.circleSelected, DEFAULT_BUTTON_COLOR))
  else:
    buttons = (("Ok", None, True, video.circleSelected, DEFAULT_BUTTON_COLOR),)
  showBlockingPage(layout, title=title, buttons=buttons, labelInfo=(frame, video, zoomable))
  return video.getInfo()


class _CollapseButton(QToolButton):
  def enterEvent(self, evt):
    QApplication.setOverrideCursor(Qt.CursorShape.ArrowCursor)
    super().enterEvent(evt)

  def leaveEvent(self, evt):
    QApplication.restoreOverrideCursor()
    super().leaveEvent(evt)


class CollapsibleSplitter(QSplitter):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.setHandleWidth(12)
    self._lastState = None

  def _collapseButtonClicked(self):
    self.setCollapsible(0, True)
    leftWidget = self.widget(0)
    orientation = self.orientation()
    if all(self.sizes()):
      self._lastState = self.saveState()
      self._collapseButton.setArrowType(Qt.ArrowType.RightArrow if orientation == Qt.Orientation.Horizontal else Qt.ArrowType.DownArrow)
      self.setSizes([0, 1])
    else:
      self._collapseButton.setArrowType(Qt.ArrowType.LeftArrow if orientation == Qt.Orientation.Horizontal else Qt.ArrowType.UpArrow)
      self.setSizes([1, 1])
      if self._lastState is not None:
        self.restoreState(self._lastState)
        self._lastState = None
    self.setCollapsible(0, False)
    self.splitterMoved.connect(self._splitterMoved)

  def _splitterMoved(self, pos, idx):
    assert idx == 1
    orientation = self.orientation()
    if pos and self._lastState is not None:
      self._collapseButton.setArrowType(Qt.ArrowType.LeftArrow if orientation == Qt.Orientation.Horizontal else Qt.ArrowType.UpArrow)
      self._lastState = None

  def createHandle(self):
    count = self.count()
    orientation = self.orientation()
    handle = QSplitterHandle(orientation, self)

    layout = QVBoxLayout() if orientation == Qt.Orientation.Horizontal else QHBoxLayout()
    layout.setContentsMargins(0, 0, 0, 0)

    self._collapseButton = _CollapseButton()
    self._collapseButton.setArrowType(Qt.ArrowType.LeftArrow if orientation == Qt.Orientation.Horizontal else Qt.ArrowType.UpArrow)
    self._collapseButton.clicked.connect(self._collapseButtonClicked)
    layout.addWidget(self._collapseButton)

    handle.setLayout(layout)
    return handle


def improveContrast(frame, quartile):
  frame = 255 - frame
  frameIsBGR = len(frame.shape) == 3
  if frameIsBGR:
    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
  lowVal = int(np.quantile(frame, quartile))
  highVal = int(np.quantile(frame, 1 - quartile))
  frame[frame < lowVal] = lowVal
  frame[frame > highVal] = highVal
  frame = frame - lowVal
  mult = np.max(frame)
  frame = frame * (255 / mult)
  frame = frame.astype('uint8')
  if frameIsBGR:
    frame = cv2.cvtColor(frame, cv2.COLOR_GRAY2BGR)
  return frame


class _RotationAngleLabel(QLabel):
  def __init__(self):
    super().__init__()
    self._currentPosition = None
    self.setMouseTracking(True)

  def mouseMoveEvent(self, evt):
    self._currentPosition = evt.pos()
    self.update()

  def enterEvent(self, evt):
    QApplication.setOverrideCursor(Qt.CursorShape.CrossCursor)

  def leaveEvent(self, evt):
    QApplication.restoreOverrideCursor()
    self._currentPosition = None
    self.update()

  def paintEvent(self, evt):
    super().paintEvent(evt)
    if self._currentPosition is None:
      return
    qp = QPainter()
    qp.begin(self)
    qp.setPen(QColor(255, 0, 0))
    qp.drawLine(0, self._currentPosition.y(), self.width(), self._currentPosition.y())
    qp.drawLine(self._currentPosition.x(), 0, self._currentPosition.x(), self.height())
    qp.end()


def getRotationAngle(frame, angle, dialog=False):
  layout = QVBoxLayout()
  video = _RotationAngleLabel()
  layout.addWidget(video, alignment=Qt.AlignmentFlag.AlignCenter, stretch=1)

  frameSlider = SliderWithSpinbox(angle, -180, 180, name="Rotation angle (degrees)", double=True)

  def getFrame():
    nonlocal angle
    angle = frameSlider.value()
    return cv2.warpAffine(frame, cv2.getRotationMatrix2D(tuple(x / 2 for x in frame.shape[1::-1]), angle, 1.0), frame.shape[1::-1], flags=cv2.INTER_LINEAR)
  frameSlider.valueChanged.connect(lambda: setPixmapFromCv(getFrame(), video))
  layout.addWidget(frameSlider, alignment=Qt.AlignmentFlag.AlignCenter)

  cancelled = False
  def cancel():
    nonlocal cancelled
    cancelled = True
  buttons = (("Cancel", cancel, True), ("Ok", None, True))
  if not dialog:
    showBlockingPage(layout, title="Select the video rotation angle", buttons=buttons, labelInfo=(getFrame(), video), dialog=dialog)
  else:
    showDialog(layout, title="Select the video rotation angle", buttons=buttons, labelInfo=(getFrame(), video))
  return None if cancelled else angle



import zebrazoom.code.util as util
import numpy as np
import math
import cv2

def computeHeading2(initialContour, lenX, lenY, headPosition, hyperparameters):
  
  xmin = lenX
  ymin = lenY
  xmax = 0
  ymax = 0
  for pt in initialContour:
    if pt[0][0] < xmin:
      xmin = pt[0][0]
    if pt[0][1] < ymin:
      ymin = pt[0][1]
    if pt[0][0] > xmax:
      xmax = pt[0][0]
    if pt[0][1] > ymax:
      ymax = pt[0][1]

  for pt in initialContour:
    pt[0][0] = pt[0][0] - xmin
    pt[0][1] = pt[0][1] - ymin
  
  headPosition = [headPosition[0] - xmin, headPosition[1] - ymin]
  
  image = np.zeros((ymax - ymin, xmax - xmin))
  image[:, :] = 255
  image = image.astype(np.uint8)
  kernel  = np.ones((3, 3), np.uint8)
  if type(initialContour) != int:
    cv2.fillPoly(image, pts =[initialContour], color=(0))
    image[:,0] = 255
    image[0,:] = 255
    image[:, len(image[0])-1] = 255
    image[len(image)-1, :]    = 255
  
  originalShape = 255 - image
  
  # Heading calculation: first approximation
  minWhitePixel = 1000000000
  bestAngle     = 0
  nTries        = 50
  for i in range(0, nTries):
    angleOption = i * ((2 * math.pi) / nTries)
    startPoint = (int(headPosition[0]), int(headPosition[1]))
    endPoint   = (int(headPosition[0] + 100000 * math.cos(angleOption)), int(headPosition[1] + 100000 * math.sin(angleOption)))
    testImage  = originalShape.copy()
    testImage  = cv2.line(testImage, startPoint, endPoint, (0), 1)
    nbWhitePixels = cv2.countNonZero(testImage)
    
    if nbWhitePixels < minWhitePixel:
      minWhitePixel = nbWhitePixels
      bestAngle     = angleOption
  bestAngleAfterFirstStep = bestAngle
  
  # Heading calculation: second (and refined) approximation
  # Searching for the optimal value of iterationsForErodeImageForHeadingCalculation
  countTries = 0
  nbIterations2nbWhitePixels = {}
  if "iterationsForErodeImageForHeadingCalculation" in hyperparameters:
    iterationsForErodeImageForHeadingCalculation = hyperparameters["iterationsForErodeImageForHeadingCalculation"]
  else:
    iterationsForErodeImageForHeadingCalculation = 4
  kernel = np.ones((3, 3), np.uint8)
  nbWhitePixelsMax = 75
  while (iterationsForErodeImageForHeadingCalculation > 0) and (countTries < 50) and not(iterationsForErodeImageForHeadingCalculation in nbIterations2nbWhitePixels):
    testImage2 = cv2.erode(testImage, kernel, iterations = iterationsForErodeImageForHeadingCalculation)
    nbWhitePixels = cv2.countNonZero(testImage2)
    nbIterations2nbWhitePixels[iterationsForErodeImageForHeadingCalculation] = nbWhitePixels
    if nbWhitePixels < nbWhitePixelsMax:
      iterationsForErodeImageForHeadingCalculation = iterationsForErodeImageForHeadingCalculation - 1
    if nbWhitePixels >= nbWhitePixelsMax:
      iterationsForErodeImageForHeadingCalculation = iterationsForErodeImageForHeadingCalculation + 1
    countTries = countTries + 1
  best_iterations = 0
  minDist = 10000000000000
  for iterations in nbIterations2nbWhitePixels:
    nbWhitePixels = nbIterations2nbWhitePixels[iterations]
    dist = abs(nbWhitePixels - nbWhitePixelsMax)
    if dist < minDist:
      minDist = dist
      best_iterations = iterations
  iterationsForErodeImageForHeadingCalculation = best_iterations
  hyperparameters["iterationsForErodeImageForHeadingCalculation"] = iterationsForErodeImageForHeadingCalculation
  
  testImage2 = cv2.erode(originalShape.copy(), kernel, iterations = iterationsForErodeImageForHeadingCalculation)
  
  maxDist = -1
  for i in range(0, nTries):
    angleOption = bestAngleAfterFirstStep - (math.pi / 5) + i * ((2 * (math.pi / 5)) / nTries)
    
    startPoint = (int(headPosition[0]), int(headPosition[1]))
    endPoint   = (int(headPosition[0] + 100000 * math.cos(angleOption)), int(headPosition[1] + 100000 * math.sin(angleOption)))
    
    testImage = testImage2.copy()
    
    testImage  = cv2.line(testImage, startPoint, endPoint, (0), 1)
    nbWhitePixels = cv2.countNonZero(testImage)
    if nbWhitePixels < minWhitePixel:
      minWhitePixel = nbWhitePixels
      bestAngle     = angleOption
    
  theta = bestAngle
  
  if hyperparameters["debugHeadingCalculation"]:
    img2 = image.copy()
    img2 = cv2.cvtColor(img2, cv2.COLOR_GRAY2RGB)
    cv2.line(img2, (int(len(img2[0])/2), int(len(img2)/2)), (int(len(img2[0])/2 + 20 * math.cos(theta)), int(len(img2)/2 + 20 * math.sin(theta))), (255,0,255), 1)
    util.showFrame(img2, title='imgForHeadingCalculation')
  
  return theta + math.pi
  
# def computeHeading(thresh1, x, y, hyperparameters):
  
  # videoWidth  = hyperparameters["videoWidth"]
  # videoHeight = hyperparameters["videoHeight"]
  # headSize = hyperparameters["headSize"]
  # ymin  = y - headSize - 10 if y - headSize >= 0 else 0
  # ymax  = y + headSize + 10 if y + headSize < len(thresh1) else len(thresh1) - 1
  # xmin  = x - headSize - 10 if x - headSize >= 0 else 0
  # xmax  = x + headSize + 10 if x + headSize < len(thresh1[0]) else len(thresh1[0]) - 1
  # img = thresh1[int(ymin):int(ymax), int(xmin):int(xmax)]
  
  # img[0,:] = 255
  # img[len(img)-1,:] = 255
  # img[:,0] = 255
  # img[:,len(img[0])-1] = 255
  
  # y2, x2 = np.nonzero(img)
  # x2 = x2 - np.mean(x2)
  # y2 = y2 - np.mean(y2)
  # coords = np.vstack([x2, y2])
  # cov = np.cov(coords)
  # evals, evecs = np.linalg.eig(cov)
  # sort_indices = np.argsort(evals)[::-1]
  # x_v1, y_v1 = evecs[:, sort_indices[0]]  # Eigenvector with largest eigenvalue
  # x_v2, y_v2 = evecs[:, sort_indices[1]]
  # scale = 20
  # theta = calculateAngle(0, 0, x_v1, y_v1)
  # theta = (theta - math.pi/2) % (2 * math.pi)
  
  # if False:
    # width  = len(img[0])
    # height = len(img)
    # option1X = x + width * math.cos(theta)
    # option1Y = y + height * math.sin(theta)
    # option2X = x - width * math.cos(theta)
    # option2Y = y - height * math.sin(theta)
    # if math.sqrt((option1X - width)**2 + (option1Y - height)**2) < math.sqrt((option2X - width)**2 + (option2Y - height)**2):
      # theta += theta + math.pi
  
  # if hyperparameters["debugHeadingCalculation"]:
    # img2 = img.copy()
    # img2 = cv2.cvtColor(img2, cv2.COLOR_GRAY2BGR)
    # cv2.line(img2, (int(len(img2[0])/2), int(len(img2)/2)), (int(len(img[0])/2 + 20 * math.cos(theta)), int(len(img)/2 + 20 * math.sin(theta))), (255,0,0), 1)
    # util.showFrame(img2, title='imgForHeadingCalculation')
  
  # return theta


import h5py
import numpy as np
import cv2
import math
import json
import sys
from scipy import interpolate
from zebrazoom.code.getImage.getForegroundImage import getForegroundImage
from zebrazoom.code.getImage.headEmbededFrame import headEmbededFrame
from scipy.interpolate import UnivariateSpline
from numpy import linspace

def Rotate(boundary, aaa, bbb, angle, dst):

  gauche = 0
  haut   = 0

  x1 = aaa + 100*math.cos(angle)
  y1 = bbb + 100*math.sin(angle)
  x2 = aaa + 100*math.cos(angle + math.pi)
  y2 = bbb + 100*math.sin(angle + math.pi)
  x = 0
  y = 0
  r = 0
  Yoo = [x1 + gauche,y1 + haut]
  Yaa = [x2 + gauche,y2 + haut]

  dist1 = 0
  min_dist1 = 1000000
  dist2 = 0
  min_dist2 = 1000000
  theta = 0
  alpha = 0
  alpha_aux = 0
  final_angle = 0
  for i in range(0, len(boundary)):
    Pt = boundary[i][0]
    dist1 = (Pt[0] - x1)*(Pt[0] - x1) + (Pt[1] - y1)*(Pt[1] - y1)
    dist2 = (Pt[0] - x2)*(Pt[0] - x2) + (Pt[1] - y2)*(Pt[1] - y2)
    if (dist1<min_dist1):
      min_dist1 = dist1
    if (dist2<min_dist2):
      min_dist2 = dist2

  if (min_dist1<min_dist2):
    theta = angle
  else:
    theta = angle + math.pi

  theta = (math.pi/2) - theta

  for i in range(0, len(boundary)):
    Pt = boundary[i][0]
    x = Pt[0]
    y = Pt[1]
    x = x - aaa
    y = y - bbb
    r = math.sqrt(x*x + y*y)
    if (x>0):
      alpha = math.atan(y/x)
    if (x<0):
      x = -x
      alpha_aux = math.atan(y/x)
      alpha = math.pi - alpha_aux
    if (x == 0):
      if (y>0):
        alpha = math.pi/2
      else:
        alpha = -math.pi/2

    final_angle = theta + alpha
    x = r*math.cos(final_angle)
    y = r*math.sin(final_angle)
    Pt[0] = x + aaa
    Pt[1] = y + bbb + 200
    
    boundary[i] = Pt
    
  return boundary



import math
import cv2
import numpy as np
from zebrazoom.code.trackingFolder.headTrackingHeadingCalculationFolder.calculateHeading import calculateHeadingSimple, calculateHeading

def simpleOptimalValueSearch(PtClosest, contour, unitVector, lenX, lenY):
  
  factor = 0
  dist = 1
  maxDist = 0
  indMax = 0
  testCenter = PtClosest + factor * unitVector
  while (dist > 0) and (factor < 20) and (testCenter[0] >= 0) and (testCenter[1] >= 0) and (testCenter[0] < lenX) and (testCenter[1] < lenY):
    factor = factor + 1
    testCenter = PtClosest + factor * unitVector
    testCenter = testCenter.astype(int)
    dist = cv2.pointPolygonTest(contour, (testCenter[0], testCenter[1]), True)
    if dist > maxDist:
      maxDist = dist
      indMax  = factor
  
  testCenter = PtClosest + indMax * unitVector
  testCenter = testCenter.astype(int)
  
  return testCenter
  

def reajustCenterOfMassIfNecessary(contour, x, y, lenX, lenY):
  inside = cv2.pointPolygonTest(contour, (x, y), True)
  if inside < 0:
    
    minDist = 100000000000000
    indMin  = 0
    for i in range(0, len(contour)):
      Pt = contour[i][0]
      dist = math.sqrt((Pt[0] - x)**2 + (Pt[1] - y)**2)
      if dist < minDist:
        minDist = dist
        indMin  = i
    PtClosest = contour[indMin][0]
    unitVector = np.array([PtClosest[0] - x, PtClosest[1] - y])
    unitVectorLength = math.sqrt(unitVector[0]**2 + unitVector[1]**2)
    unitVector[0] = unitVector[0] / unitVectorLength
    unitVector[1] = unitVector[1] / unitVectorLength
    if True:
      factor = 5
      testCenter = PtClosest + factor * unitVector
      testCenter = testCenter.astype(int)
      while (cv2.pointPolygonTest(contour, (testCenter[0], testCenter[1]), True) <= 0) and (factor > 1):
        factor = factor - 1
        testCenter = PtClosest + factor * unitVector
    else:
      testCenter = simpleOptimalValueSearch(PtClosest, contour, unitVector, lenX, lenY)
    
    x = testCenter[0]
    y = testCenter[1]
  
  return [x, y]
  
def findCenterByIterativelyDilating(initialContour, lenX, lenY):
  x = 0
  y = 0
  image = np.zeros((lenY, lenX))
  image[:, :] = 255
  image = image.astype(np.uint8)
  kernel  = np.ones((3, 3), np.uint8)
  if type(initialContour) != int:
    cv2.fillPoly(image, pts =[initialContour], color=(0))
    image[:,0] = 255
    image[0,:] = 255
    image[:, len(image[0])-1] = 255
    image[len(image)-1, :]    = 255
    nbBlackPixels = 1
    dilateIter = 0
    while nbBlackPixels > 0:
      dilateIter   = dilateIter + 1
      dilatedImage = cv2.dilate(image, kernel, iterations=dilateIter)
      nbBlackPixels = cv2.countNonZero(255-dilatedImage)
    dilateIter   = dilateIter - 1
    dilatedImage = cv2.dilate(image, kernel, iterations=dilateIter)
    dilatedImage[:,0] = 255
    dilatedImage[0,:] = 255
    dilatedImage[:, len(dilatedImage[0])-1] = 255
    dilatedImage[len(dilatedImage)-1, :]    = 255
    contours, hierarchy = cv2.findContours(dilatedImage, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
    maxContour = 0
    maxContourArea = 0
    for contour in contours:
      contourArea = cv2.contourArea(contour)
      if contourArea < int(lenX * lenY * 0.8):
        if contourArea > maxContourArea:
          maxContourArea = contourArea
          maxContour     = contour
    M = cv2.moments(maxContour)
    if M['m00']:
      x = int(M['m10']/M['m00'])
      y = int(M['m01']/M['m00'])
  return [x, y]
   

def multipleAnimalsHeadTracking(trackingHeadingAllAnimals, trackingHeadTailAllAnimals, hyperparameters, gray, i, firstFrame, thresh1, xmin=0, ymin=0):
  
  headCoordinatesOptions      = []
  headCoordinatesAreasOptions = []
  x = 0
  y = 0
  minAreaCur = hyperparameters["minArea"]
  maxAreaCur = hyperparameters["maxArea"]
  meanArea   = (hyperparameters["minArea"] + hyperparameters["maxArea"]) / 2
  
  thresh2 = thresh1.copy()
  erodeSize = hyperparameters["erodeSize"]
  if erodeSize:
    kernel  = np.ones((erodeSize,erodeSize), np.uint8)
    thresh2 = cv2.dilate(thresh2, kernel, iterations=hyperparameters["dilateIter"])
  
  thresh2[:,0] = 255
  thresh2[0,:] = 255
  thresh2[:, len(thresh2[0])-1] = 255
  thresh2[len(thresh2)-1, :]    = 255
  
  if hyperparameters["multipleHeadTrackingIterativelyRelaxAreaCriteria"]:
    contours, hierarchy = cv2.findContours(thresh2,cv2.RETR_TREE,cv2.CHAIN_APPROX_SIMPLE)
    areaDecreaseFactor = int(hyperparameters["minArea"]/10)
    while (len(headCoordinatesOptions) < hyperparameters["nbAnimalsPerWell"]) and (minAreaCur > 0):
      headCoordinatesOptions = []
      headCoordinatesAreasOptions = []
      for contour in contours:
        area = cv2.contourArea(contour)
        if (area > minAreaCur) and (area < maxAreaCur):
          if hyperparameters["findCenterOfAnimalByIterativelyDilating"]:
            [x, y] = findCenterByIterativelyDilating(contour, len(thresh2[0]), len(thresh2))
          else:
            M = cv2.moments(contour)
            if M['m00']:
              x = int(M['m10']/M['m00'])
              y = int(M['m01']/M['m00'])
            else:
              x = 0
              y = 0
            if hyperparameters["readjustCenterOfMassIfNotInsideContour"]:
              [x, y] = reajustCenterOfMassIfNecessary(contour, x, y, len(thresh2[0]), len(thresh2))
          if not([x, y] in headCoordinatesOptions):
            headCoordinatesOptions.append([x+xmin, y+ymin])
            headCoordinatesAreasOptions.append(abs(area - (meanArea/2)))
      
      if areaDecreaseFactor:
        minAreaCur = minAreaCur - areaDecreaseFactor
        maxAreaCur = int(maxAreaCur * 1.1)
      else:
        minAreaCur = minAreaCur - 1
        maxAreaCur = int(maxAreaCur * 1.1)
  else:
    contours, hierarchy = cv2.findContours(thresh2,cv2.RETR_TREE,cv2.CHAIN_APPROX_SIMPLE)
    for contour in contours:
      area = cv2.contourArea(contour)
      if (area > minAreaCur) and (area < maxAreaCur):
        # print("Find center of mass: area:", area)
        if hyperparameters["findCenterOfAnimalByIterativelyDilating"]:
          [x, y] = findCenterByIterativelyDilating(contour, len(thresh2[0]), len(thresh2))
        else:
          M = cv2.moments(contour)
          if M['m00']:
            x = int(M['m10']/M['m00'])
            y = int(M['m01']/M['m00'])
          else:
            x = 0
            y = 0
          if hyperparameters["readjustCenterOfMassIfNotInsideContour"]:
            [x, y] = reajustCenterOfMassIfNecessary(contour, x, y, len(thresh2[0]), len(thresh2))
        headCoordinatesOptions.append([x+xmin, y+ymin])
        headCoordinatesAreasOptions.append(abs(area - (meanArea/2)))

  headCoordinatesAreasOptionsSorted = np.argsort(headCoordinatesAreasOptions)
  headCoordinatesOptions = [headCoordinatesOptions[elem] for elem in headCoordinatesAreasOptionsSorted]
  
  headCoordinatesOptionsAlreadyTakenDist     = [-1 for k in headCoordinatesOptions]
  headCoordinatesOptionsAlreadyTakenAnimalId = [-1 for k in headCoordinatesOptions]
  animalNotPutOrEjectedBecausePositionAlreadyTaken = 1
  if i > firstFrame:
    while animalNotPutOrEjectedBecausePositionAlreadyTaken:
      animalNotPutOrEjectedBecausePositionAlreadyTaken = 0
      for animal_Id in range(0, hyperparameters["nbAnimalsPerWell"]):
        x_prevFrame_animal_Id = trackingHeadTailAllAnimals[animal_Id, i-firstFrame-1][0][0]
        y_prevFrame_animal_Id = trackingHeadTailAllAnimals[animal_Id, i-firstFrame-1][0][1]
        x_curFrame_animal_Id  = trackingHeadTailAllAnimals[animal_Id, i-firstFrame][0][0]
        y_curFrame_animal_Id  = trackingHeadTailAllAnimals[animal_Id, i-firstFrame][0][1]
        if ((x_prevFrame_animal_Id != 0) and (y_prevFrame_animal_Id != 0) and (x_curFrame_animal_Id == 0) and (y_curFrame_animal_Id == 0)):
          min_dist  = 10000000000000000000000
          index_min = -1
          for idxCoordinateOption, [idx_x_Option, idx_y_Option] in enumerate(headCoordinatesOptions):
            dist = (idx_x_Option - x_prevFrame_animal_Id)**2 + (idx_y_Option - y_prevFrame_animal_Id)**2
            if dist < min_dist:
              if headCoordinatesOptionsAlreadyTakenDist[idxCoordinateOption] != -1:
                if dist < headCoordinatesOptionsAlreadyTakenDist[idxCoordinateOption]:
                  min_dist = dist
                  index_min = idxCoordinateOption
              else:
                min_dist = dist
                index_min = idxCoordinateOption
          # if len(headCoordinatesOptions):
          if index_min >= 0:
            if (headCoordinatesOptionsAlreadyTakenDist[index_min] == -1):
              # Save position of animal for frame i-firstFrame
              trackingHeadTailAllAnimals[animal_Id, i-firstFrame][0][0] = headCoordinatesOptions[index_min][0]
              trackingHeadTailAllAnimals[animal_Id, i-firstFrame][0][1] = headCoordinatesOptions[index_min][1]     
              headCoordinatesOptionsAlreadyTakenDist[index_min]     = min_dist
              headCoordinatesOptionsAlreadyTakenAnimalId[index_min] = animal_Id
            else:
              animalNotPutOrEjectedBecausePositionAlreadyTaken = 1
              if min_dist < headCoordinatesOptionsAlreadyTakenDist[index_min]:
                # "Ejecting" previously saved position of animal for frame i-firstFrame
                animal_Id_Eject = headCoordinatesOptionsAlreadyTakenAnimalId[index_min]
                trackingHeadTailAllAnimals[animal_Id_Eject, i-firstFrame][0][0] = 0
                trackingHeadTailAllAnimals[animal_Id_Eject, i-firstFrame][0][1] = 0          
                # Replace position of animal for frame i-firstFrame
                trackingHeadTailAllAnimals[animal_Id, i-firstFrame][0][0] = headCoordinatesOptions[index_min][0]
                trackingHeadTailAllAnimals[animal_Id, i-firstFrame][0][1] = headCoordinatesOptions[index_min][1]  
                headCoordinatesOptionsAlreadyTakenDist[index_min]     = min_dist
                headCoordinatesOptionsAlreadyTakenAnimalId[index_min] = animal_Id               
  
  headCoordinatesOptions = [headCoordinatesOptions[idx] for idx, k in enumerate(headCoordinatesOptionsAlreadyTakenAnimalId) if k == -1]
  
  for idxCoordinateOption, [idx_x_Option, idx_y_Option] in enumerate(headCoordinatesOptions):
    for animal_Id in range(0, hyperparameters["nbAnimalsPerWell"]):
      if (trackingHeadTailAllAnimals[animal_Id, i-firstFrame][0][0] == 0) and (trackingHeadTailAllAnimals[animal_Id, i-firstFrame][0][1] == 0):
        if i == firstFrame:
          trackingHeadTailAllAnimals[animal_Id, i-firstFrame][0][0] = idx_x_Option
          trackingHeadTailAllAnimals[animal_Id, i-firstFrame][0][1] = idx_y_Option
          break
        else:
          if (trackingHeadTailAllAnimals[animal_Id, i-firstFrame-1][0][0] == 0) and (trackingHeadTailAllAnimals[animal_Id, i-firstFrame-1][0][1] == 0): # NEW !!!!
            trackingHeadTailAllAnimals[animal_Id, i-firstFrame][0][0] = idx_x_Option
            trackingHeadTailAllAnimals[animal_Id, i-firstFrame][0][1] = idx_y_Option
            break

  if hyperparameters["postProcessMultipleTrajectories"]:
    maxDistanceAuthorized = hyperparameters["postProcessMaxDistanceAuthorized"]
    maxDisapearanceFrames = hyperparameters["postProcessMaxDisapearanceFrames"]
    for animal_Id in range(0, len(trackingHeadTailAllAnimals)):
      goBackFrames = 1
      xCur = trackingHeadTailAllAnimals[animal_Id, i-firstFrame][0][0]
      yCur = trackingHeadTailAllAnimals[animal_Id, i-firstFrame][0][1]
      if xCur != 0 or yCur != 0:
        while (i-firstFrame-goBackFrames >= 0) and (goBackFrames < maxDisapearanceFrames) and (trackingHeadTailAllAnimals[animal_Id, i-firstFrame-goBackFrames][0][0] == 0) and (trackingHeadTailAllAnimals[animal_Id, i-firstFrame-goBackFrames][0][1] == 0):
          goBackFrames = goBackFrames + 1
        if (goBackFrames < maxDisapearanceFrames) and (i-firstFrame-goBackFrames >= 0):
          xBef = trackingHeadTailAllAnimals[animal_Id, i-firstFrame-goBackFrames][0][0]
          yBef = trackingHeadTailAllAnimals[animal_Id, i-firstFrame-goBackFrames][0][1]
          if math.sqrt((xCur - xBef)**2 + (yCur - yBef)**2) > maxDistanceAuthorized:
            trackingHeadTailAllAnimals[animal_Id, i-firstFrame][0][0] = 0
            trackingHeadTailAllAnimals[animal_Id, i-firstFrame][0][1] = 0
  
  # for idxCoordinateOption, [idx_x_Option, idx_y_Option] in enumerate(headCoordinatesOptions):
    # print(i, idxCoordinateOption, idx_x_Option, idx_y_Option)
    # min_dist  = 10000000000000000000000
    # index_min_animal_Id = -1
    # maxdepth_min_animal_Id = 0
    # index_animal_Id_alwaysBeenAt0 = -1
    # for animal_Id in range(0, hyperparameters["nbAnimalsPerWell"]):
      # if (trackingHeadTailAllAnimals[animal_Id, i-firstFrame][0][0] == 0) and (trackingHeadTailAllAnimals[animal_Id, i-firstFrame][0][1] == 0):
        # depth = 1
        # while (i - firstFrame - depth >= 0) and (depth < 20) and (trackingHeadTailAllAnimals[animal_Id, i-firstFrame-depth][0][0] == 0) and (trackingHeadTailAllAnimals[animal_Id, i-firstFrame-depth][0][1] == 0):
          # depth = depth + 1
        # xPast_animal_Id = trackingHeadTailAllAnimals[animal_Id, i-firstFrame-depth][0][0]
        # yPast_animal_Id = trackingHeadTailAllAnimals[animal_Id, i-firstFrame-depth][0][1]
        # if (xPast_animal_Id != 0) or (yPast_animal_Id != 0):
          # print("animal_Id:", animal_Id, "; (xPast_animal_Id, yPast_animal_Id):", xPast_animal_Id, yPast_animal_Id)
          # dist = math.sqrt((idx_x_Option - xPast_animal_Id)**2 + (idx_y_Option - yPast_animal_Id)**2)
          # print("dist:", dist)
          # if dist < 30:
            # print("depth", depth)
            # if dist < min_dist:
              # min_dist = dist
              # index_min_animal_Id = animal_Id
              # maxdepth_min_animal_Id = depth
        # else:
          # if index_animal_Id_alwaysBeenAt0 == -1:
            # index_animal_Id_alwaysBeenAt0 = animal_Id
    # print("index_min_animal_Id:", index_min_animal_Id)
    # if index_min_animal_Id >= 0:
      # trackingHeadTailAllAnimals[index_min_animal_Id, i-firstFrame][0][0] = idx_x_Option
      # trackingHeadTailAllAnimals[index_min_animal_Id, i-firstFrame][0][1] = idx_y_Option
      # x_maxPast_animal_Id = trackingHeadTailAllAnimals[index_min_animal_Id, i-firstFrame-maxdepth_min_animal_Id][0][0]
      # y_maxPast_animal_Id = trackingHeadTailAllAnimals[index_min_animal_Id, i-firstFrame-maxdepth_min_animal_Id][0][1]
      # print("x_maxPast_animal_Id, y_maxPast_animal_Id:", x_maxPast_animal_Id, y_maxPast_animal_Id)
      # print("maxdepth_min_animal_Id", maxdepth_min_animal_Id)
      # x_step = (idx_x_Option - x_maxPast_animal_Id) / maxdepth_min_animal_Id
      # y_step = (idx_y_Option - y_maxPast_animal_Id) / maxdepth_min_animal_Id
      # for blankSpace in range(1, maxdepth_min_animal_Id):
        # trackingHeadTailAllAnimals[index_min_animal_Id, i-firstFrame-maxdepth_min_animal_Id+blankSpace][0][0] = x_maxPast_animal_Id + blankSpace * x_step
        # trackingHeadTailAllAnimals[index_min_animal_Id, i-firstFrame-maxdepth_min_animal_Id+blankSpace][0][1] = y_maxPast_animal_Id + blankSpace * y_step
    # else:
      # if index_animal_Id_alwaysBeenAt0 != -1:
        # trackingHeadTailAllAnimals[index_animal_Id_alwaysBeenAt0, i-firstFrame][0][0] = idx_x_Option
        # trackingHeadTailAllAnimals[index_animal_Id_alwaysBeenAt0, i-firstFrame][0][1] = idx_y_Option
  
  if hyperparameters["findCenterOfAnimalByIterativelyDilating"] == 0 or hyperparameters["trackTail"] == 0:
    for animal_Id in range(0, hyperparameters["nbAnimalsPerWell"]):
      x_curFrame_animal_Id  = trackingHeadTailAllAnimals[animal_Id, i-firstFrame][0][0]
      y_curFrame_animal_Id  = trackingHeadTailAllAnimals[animal_Id, i-firstFrame][0][1]
      if x_curFrame_animal_Id != 0 and y_curFrame_animal_Id != 0:
        x_curFrame_animal_Id = int(x_curFrame_animal_Id)
        y_curFrame_animal_Id = int(y_curFrame_animal_Id)
        
        # Removing the other blobs from the image to get good heading detection
        # Performance improvement possible below: TODO in the future
        correspondingContour = 0
        contours, hierarchy = cv2.findContours(thresh2,cv2.RETR_TREE,cv2.CHAIN_APPROX_SIMPLE)
        for contour in contours:
          dist = cv2.pointPolygonTest(contour, (x_curFrame_animal_Id, y_curFrame_animal_Id), True)
          if dist >= 0:
            correspondingContour = contour
        thresh2bis = np.zeros((len(thresh2), len(thresh2[0])))
        thresh2bis[:, :] = 255
        if type(correspondingContour) != int:
          cv2.fillPoly(thresh2bis, pts =[correspondingContour], color=(0, 0, 0))
        
        if hyperparameters["forceBlobMethodForHeadTracking"]: # Fish tracking, usually
          previousFrameHeading = trackingHeadingAllAnimals[animal_Id, i-firstFrame-1] if i-firstFrame-1 > 0 else 0
          [heading, lastFirstTheta] = calculateHeading(x_curFrame_animal_Id, y_curFrame_animal_Id, 0, thresh1, thresh2bis, 0, hyperparameters, previousFrameHeading)
        else: # Drosophilia, mouse, and other center of mass only tracking
          heading = calculateHeadingSimple(x_curFrame_animal_Id, y_curFrame_animal_Id, thresh2bis, hyperparameters)
        
        trackingHeadingAllAnimals[animal_Id, i-firstFrame] = heading
  
  return [trackingHeadingAllAnimals, trackingHeadTailAllAnimals]



from zebrazoom.code.preprocessImage import preprocessImage
import numpy as np
import cv2
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading

def headEmbededFrame(videoPath, frameNumber, wellNumber, wellPositions, hyperparameters):
  
  debug = 0
  
  cap = zzVideoReading.VideoCapture(videoPath)
  
  cap.set(1, frameNumber)
  ret, frame = cap.read()
  
  while not(ret):
    print("WARNING: couldn't read the frameNumber", frameNumber, "for the video", hyperparameters["videoName"])
    frameNumber = frameNumber - 1
    cap.set(1, frameNumber)
    ret, frame = cap.read()

  xtop = wellPositions[wellNumber]['topLeftX']
  ytop = wellPositions[wellNumber]['topLeftY']
  lenX = wellPositions[wellNumber]['lengthX']
  lenY = wellPositions[wellNumber]['lengthY']
  frame = frame[ytop:ytop+lenY, xtop:xtop+lenX]

  if ("invertBlackWhiteOnImages" in hyperparameters) and hyperparameters["invertBlackWhiteOnImages"]:
    frame = 255 - frame
  
  if ("imagePreProcessMethod" in hyperparameters) and hyperparameters["imagePreProcessMethod"]:
    frame = preprocessImage(frame, hyperparameters)
  
  kernel = np.ones((8,8),np.float32)/25
  thres1  = cv2.filter2D(frame,-1,kernel)
  retval, thres1 = cv2.threshold(thres1, 80, 255, cv2.THRESH_BINARY)
  thres1 = 255 - thres1
  
  frame = 255 - frame
  
  if (debug):
    import zebrazoom.code.util as util
    util.showFrame(frame, title='thres1')
    util.showFrame(thres1, title='thres1')
    
  frame  = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
  thres1 = cv2.cvtColor(thres1, cv2.COLOR_BGR2GRAY)
  
  return [frame, thres1]



import h5py
import numpy as np
import cv2
import math
import json
import sys
from scipy import interpolate
from zebrazoom.code.getImage.getForegroundImage import getForegroundImage
from zebrazoom.code.getImage.headEmbededFrame import headEmbededFrame
from scipy.interpolate import UnivariateSpline
from numpy import linspace

from zebrazoom.code.trackingFolder.tailTrackingFunctionsFolder.tailTrackingExtremityDetectFolder.resampleSeqConstPtsPerArcLength import resampleSeqConstPtsPerArcLength

def fillTailRanges(tailRange1,tailRange2,fillSecond,i,MostCurvyIndex):
  if (i == MostCurvyIndex):
    fillSecond = 1
  if fillSecond == 0:
    tailRange1.append(i)
  else:
    tailRange2.append(i)
  return [tailRange1,tailRange2,fillSecond]

def getMidline(bord1, bord2, MostCurvyIndex, boundary, dst, taille, distance2, debug, hyperparameters, nbTailPoints):

  nbTailPoints = taille
  output = np.zeros((1, 0, 2))

  numTailPoints = nbTailPoints
  minTailSize = 20
  maxTailSize = 60
  trackingPointSizeDisplay = 1

  OrigBoundA = []
  OrigBoundB = []
  
  if (bord2 < bord1):
    temp  = bord2
    bord2 = bord1
    bord1 = temp
    
  max1 = distance2[bord2] - distance2[bord1] 
  max2 = (distance2[bord1] - distance2[0])  + (distance2[len(boundary)] - distance2[bord2])
  
  tailRangeA = []
  tailRangeB = []
  fillSecond = 0
  if (max1 > max2):
    for i in range(bord1, bord2):
      [tailRangeA,tailRangeB,fillSecond] = fillTailRanges(tailRangeA,tailRangeB,fillSecond,i,MostCurvyIndex)
  else:
    for i in range(bord2, len(boundary)):
      [tailRangeA,tailRangeB,fillSecond] = fillTailRanges(tailRangeA,tailRangeB,fillSecond,i,MostCurvyIndex)
    for i in range(0, bord1):
      [tailRangeA,tailRangeB,fillSecond] = fillTailRanges(tailRangeA,tailRangeB,fillSecond,i,MostCurvyIndex)
  
  OrigBoundA = boundary[tailRangeA]
  OrigBoundB = boundary[tailRangeB]
  
  if ((bord1!=bord2) and (bord1!=MostCurvyIndex) and (bord2!=MostCurvyIndex) and not((bord1==1) and (bord2==1) and (MostCurvyIndex==1)) and (len(OrigBoundA)>1) and (len(OrigBoundB)>1)):
    
    if False:
      for pt in OrigBoundA:
        cv2.circle(dst, (pt[0][0], pt[0][1]), 1, (0, 255, 0), -1)
      for pt in OrigBoundB:
        cv2.circle(dst, (pt[0][0], pt[0][1]), 1, (255, 0, 0), -1)
      import zebrazoom.code.util as util

      util.showFrame(dst, title='dst')
    
    NBoundA = resampleSeqConstPtsPerArcLength(OrigBoundA, numTailPoints)
    NBoundB = resampleSeqConstPtsPerArcLength(OrigBoundB, numTailPoints)

    # calculates length of the tail
    TotalDist = 0
    for i in range(1, taille):
      Pt  = NBoundB[i % taille]
      Pt2 = NBoundA[taille - i]
      x = (Pt[0]+Pt2[0]) / 2
      y = (Pt[1]+Pt2[1]) / 2
      if i > 1:
        TotalDist = TotalDist + math.sqrt((x-xAvant)*(x-xAvant)+(y-yAvant)*(y-yAvant))
      xAvant = x
      yAvant = y
      
    if ((TotalDist<hyperparameters["minTailSize"]) or (TotalDist>hyperparameters["maxTailSize"])):
    
      if (debug):
        print("innapropriate tail size! TailDist: ", TotalDist, " ; but minTailSize is ", minTailSize, " and maxTailSize is ", maxTailSize)
        
    else:
      
      Tail = boundary[MostCurvyIndex][0]
      
      point = np.array([Tail[0], Tail[1]])
      output = np.insert(output, 0, point, axis=1)
      
      for i in range(1, taille):
        Pt  = NBoundB[i % taille]
        Pt2 = NBoundA[taille - i]
        point = np.array([(Pt[0]+Pt2[0])/2, (Pt[1]+Pt2[1])/2])
        output = np.insert(output, 0, point, axis=1)
      
      i = taille-2
      if i >= 1:
        Pt =  NBoundB[i % taille]
        Pt2 = NBoundA[taille-i]
        ClosestPoint = [ (Pt[0]+Pt2[0])/2 , (Pt[1]+Pt2[1])/2 ]
      else:
        ClosestPoint = [-200, -200]
  
  else:
  
    # THIS SHOULD BE IMPROVED IN THE FUTURE:
    # WE SHOULD CHECK FOR TAIL LENGHT
    # ALSO WE SHOULD DO SOMETHING BETTER THAN JUST PUTTING THE TAIL TIP FOR EACH OF THE TEN POINTS !!!
    Tail = boundary[MostCurvyIndex][0]    
    point = np.array([Tail[0], Tail[1]])
    for i in range(0, taille):
      output = np.insert(output, 0, point, axis=1)    
  
  return output



import cv2

def detectMovementWithRawVideoInsideTracking2(hyperparameters, trackingHeadTailAllAnimalsList, previousFrames, animal_Id, i, firstFrame, auDessusPerAnimalIdList, grey, wellPositions):
  halfDiameterRoiBoutDetect = hyperparameters["halfDiameterRoiBoutDetect"]
  if previousFrames.full():
    previousFrame   = previousFrames.get()
    curFrame        = grey.copy()
    for wellNumber in range(0 if hyperparameters["onlyTrackThisOneWell"] == -1 else hyperparameters["onlyTrackThisOneWell"], hyperparameters["nbWells"] if hyperparameters["onlyTrackThisOneWell"] == -1 else hyperparameters["onlyTrackThisOneWell"] + 1):
      # previousXYCoord = previousXYCoords.get()
      headX = trackingHeadTailAllAnimalsList[wellNumber][animal_Id, i-firstFrame][0][0]
      headY = trackingHeadTailAllAnimalsList[wellNumber][animal_Id, i-firstFrame][0][1]
      xmin = headX - hyperparameters["halfDiameterRoiBoutDetect"]
      ymin = headY - hyperparameters["halfDiameterRoiBoutDetect"]
      xmax = xmin + 2 * hyperparameters["halfDiameterRoiBoutDetect"]
      ymax = ymin + 2 * hyperparameters["halfDiameterRoiBoutDetect"]
      lenX = wellPositions[wellNumber]['lengthX']
      lenY = wellPositions[wellNumber]['lengthY']
      if xmin < 0:
        xmin = 0
      if ymin < 0:
        ymin = 0
      if xmax > lenX - 1:
        xmax = lenX - 1
      if ymax > lenY - 1:
        ymax = lenY - 1
      if ymax < ymin:
        ymax = ymin + 2 * hyperparameters["halfDiameterRoiBoutDetect"]
      if xmax < xmin:
        xmax = xmin + 2 * hyperparameters["halfDiameterRoiBoutDetect"]
      if ( (xmin > lenX - 1) or (xmax < 0) ):
        xmin = 0
        xmax = 0 + lenX - 1
      if ( (ymin > lenY - 1) or (ymax < 0) ):
        ymin = 0
        ymax = 0 + lenY - 1
      xmin = int(xmin + wellPositions[wellNumber]['topLeftX'])
      xmax = int(xmax + wellPositions[wellNumber]['topLeftX'])
      ymin = int(ymin + wellPositions[wellNumber]['topLeftY'])
      ymax = int(ymax + wellPositions[wellNumber]['topLeftY'])
      # img22       = img2[ymin:ymax, xmin:xmax]
      # imgFuture22 = imgFuture2[ymin:ymax, xmin:xmax]
      # maxX = min(len(previousFrame[0]), len(curFrame[0]))
      # maxY = min(len(previousFrame), len(curFrame))
      # print("Av: aaa:", len(previousFrame))
      # print("Av: bbb:", len(previousFrame[0]))
      # print("1Av: aaa:", len(grey))
      # print("1Av: bbb:", len(grey[0]))
      subPreviousFrame = previousFrame[ymin:ymax, xmin:xmax].copy()
      subCurFrame      = curFrame[ymin:ymax, xmin:xmax].copy()
      # print("Aft: aaa:", len(previousFrame))
      # print("Aft: bbb:", len(previousFrame[0]))            
      # if i > 250 and wellNumber == 3:
        # import zebrazoom.code.util as util
        # print("wellNumber:", wellNumber, "; headX, headY:", headX, headY, "; xmin, xmax, ymin, ymax:", xmin, xmax, ymin, ymax)
        # print("aaa:", len(previousFrame))
        # print("bbb:", len(previousFrame[0]))
        # util.showFrame(subPreviousFrame, title='subPreviousFrame' + str(wellNumber))
        # util.showFrame(subCurFrame,      title='subCurFrame' + str(wellNumber))
      
      # Possible optimization in the future: refine the ROI based on halfDiameterRoiBoutDetect !!!
      res = cv2.absdiff(subPreviousFrame, subCurFrame)
      ret, res = cv2.threshold(res,hyperparameters["thresForDetectMovementWithRawVideo"],255,cv2.THRESH_BINARY)
      totDiff = cv2.countNonZero(res)
      
      for animalId in range(0, hyperparameters["nbAnimalsPerWell"]):
        if totDiff > hyperparameters["minNbPixelForDetectMovementWithRawVideo"]:
          auDessusPerAnimalIdList[wellNumber][animalId][i-firstFrame] = 1
        else:
          auDessusPerAnimalIdList[wellNumber][animalId][i-firstFrame] = 0
  else:
    for wellNumber in range(0 if hyperparameters["onlyTrackThisOneWell"] == -1 else hyperparameters["onlyTrackThisOneWell"], hyperparameters["nbWells"] if hyperparameters["onlyTrackThisOneWell"] == -1 else hyperparameters["onlyTrackThisOneWell"] + 1):
      for animalId in range(0, hyperparameters["nbAnimalsPerWell"]):
        auDessusPerAnimalIdList[wellNumber][animalId][i-firstFrame] = 0
  previousFrames.put(grey)
  # previousXYCoords.put([xHead, yHead])
  
  return [auDessusPerAnimalIdList, previousFrames]



import h5py
import numpy as np
import cv2
import math
import json
import sys
from scipy import interpolate
from zebrazoom.code.getImage.getForegroundImage import getForegroundImage
from zebrazoom.code.getImage.headEmbededFrame import headEmbededFrame
from zebrazoom.code.deepLearningFunctions.labellingFunctions import drawWhitePointsOnInitialImages, saveImagesAndData
from scipy.interpolate import UnivariateSpline
from numpy import linspace

def findBodyContour(headPosition, hyperparameters, thresh1, initialCurFrame, back, wellNumber=-1, frameNumber=-1):

  if hyperparameters["saveBodyMask"] and hyperparameters["bodyMask_addWhitePoints"]:
    [img, thresh1] = drawWhitePointsOnInitialImages(initialCurFrame, back, hyperparameters)
  
  thresh1[:,0] = 255
  thresh1[0,:] = 255
  thresh1[:, len(thresh1[0])-1] = 255
  thresh1[len(thresh1)-1, :]    = 255
  
  x = headPosition[0]
  y = headPosition[1]
  cx = 0
  cy = 0
  takeTheHeadClosestToTheCenter = 1
  bodyContour = 0

  if hyperparameters["findContourPrecision"] == "CHAIN_APPROX_SIMPLE":
    contourPrecision = cv2.CHAIN_APPROX_SIMPLE
  else: # hyperparameters["findContourPrecision"] == "CHAIN_APPROX_NONE"
    contourPrecision = cv2.CHAIN_APPROX_NONE
  
  if hyperparameters["recalculateForegroundImageBasedOnBodyArea"]:
    
    minPixel2nbBlackPixels = {}
    countTries = 0
    nbBlackPixels = 0
    nbBlackPixelsMax = int(hyperparameters["adjustMinPixelDiffForBackExtract_nbBlackPixelsMax"] / hyperparameters["nbAnimalsPerWell"])
    minPixelDiffForBackExtract = int(hyperparameters["minPixelDiffForBackExtract"])
    if "minPixelDiffForBackExtractBody" in hyperparameters:
      minPixelDiffForBackExtract = hyperparameters["minPixelDiffForBackExtractBody"]
    
    previousNbBlackPixels = []
    while (minPixelDiffForBackExtract > 0) and (countTries < 30) and not(minPixelDiffForBackExtract in minPixel2nbBlackPixels):
      curFrame = initialCurFrame.copy()
      putToWhite = ( curFrame.astype('int32') >= (back.astype('int32') - minPixelDiffForBackExtract) )
      curFrame[putToWhite] = 255
      ret, thresh1_b = cv2.threshold(curFrame, hyperparameters["thresholdForBlobImg"], 255, cv2.THRESH_BINARY)
      thresh1_b = 255 - thresh1_b
      bodyContour = 0
      contours, hierarchy = cv2.findContours(thresh1_b, cv2.RETR_TREE, contourPrecision)
      for contour in contours:
        dist = cv2.pointPolygonTest(contour, (x, y), True)
        if dist >= 0:
          M = cv2.moments(contour)
          if M['m00']:
            cx = int(M['m10']/M['m00'])
            cy = int(M['m01']/M['m00'])
            bodyContour = contour
          else:
            cx = 0
            cy = 0
      if not(type(bodyContour) == int):
        nbBlackPixels = cv2.contourArea(bodyContour)
      else:
        nbBlackPixels = -100000000
    
      minPixel2nbBlackPixels[minPixelDiffForBackExtract] = nbBlackPixels
      if nbBlackPixels > nbBlackPixelsMax:
        minPixelDiffForBackExtract = minPixelDiffForBackExtract + 1
      if nbBlackPixels <= nbBlackPixelsMax:
        minPixelDiffForBackExtract = minPixelDiffForBackExtract - 1
      
      countTries = countTries + 1
      
      previousNbBlackPixels.append(nbBlackPixels)
      if len(previousNbBlackPixels) >= 3:
        lastThree = previousNbBlackPixels[len(previousNbBlackPixels)-3: len(previousNbBlackPixels)]
        if lastThree.count(lastThree[0]) == len(lastThree):
          countTries = 1000000
    
    best_minPixelDiffForBackExtract = 0
    minDist = 10000000000000
    for minPixelDiffForBackExtract in minPixel2nbBlackPixels:
      nbBlackPixels = minPixel2nbBlackPixels[minPixelDiffForBackExtract]
      dist = abs(nbBlackPixels - nbBlackPixelsMax)
      if dist < minDist:
        minDist = dist
        best_minPixelDiffForBackExtract = minPixelDiffForBackExtract
        
    minPixelDiffForBackExtract = best_minPixelDiffForBackExtract
    putToWhite = (curFrame.astype('int32') >= (back.astype('int32') - minPixelDiffForBackExtract) )
    curFrame[putToWhite] = 255
    
    ret, thresh1 = cv2.threshold(curFrame, hyperparameters["thresholdForBlobImg"], 255, cv2.THRESH_BINARY)
    thresh1 = 255 - thresh1
    
    hyperparameters["minPixelDiffForBackExtractBody"] = minPixelDiffForBackExtract
  
  contours, hierarchy = cv2.findContours(thresh1, cv2.RETR_TREE, contourPrecision)
  for contour in contours:
    area = cv2.contourArea(contour)
    if (area >= hyperparameters["minAreaBody"]) and (area <= hyperparameters["maxAreaBody"]):
      dist = cv2.pointPolygonTest(contour, (x, y), True)
      if dist >= 0 or hyperparameters["saveBodyMask"]:
        M = cv2.moments(contour)
        if M['m00']:
          cx = int(M['m10']/M['m00'])
          cy = int(M['m01']/M['m00'])
          bodyContour = contour
        else:
          cx = 0
          cy = 0
  
  if type(bodyContour) != int:
    if cv2.contourArea(bodyContour) >= hyperparameters["maxAreaBody"]:
      bodyContour = 0
  
  if hyperparameters["saveBodyMask"]:
    saveImagesAndData(hyperparameters, bodyContour, initialCurFrame, wellNumber, frameNumber)
        
  return bodyContour



# WORK IN PROGRESS

import numpy as np
from scipy.interpolate import UnivariateSpline

def getInstaMaxTailBendPos(curbout, nbFramesTakenIntoAccount):

  fin = curbout["BoutEnd"] - curbout["BoutStart"] + 1
  
  if fin > 2:
    tailPosX = curbout["TailX_VideoReferential"].copy()
    tailPosY = curbout["TailY_VideoReferential"].copy()
    for i in range(0, fin):
      tailPosXcur = tailPosX[i]
      tailPosYcur = tailPosY[i]
      for j in range(1, len(tailPosXcur)-1):
        
    
    instMaxTailBendPos = 
    instMaxTailBendPos = np.append(instMaxTailBendPos, instMaxTailBendPos[len(instMaxTailBendPos)-1])
  else:
    instMaxTailBendPos = [0]*fin
    
  instMaxTailBendPos2  = [0]*nbFramesTakenIntoAccount
  
  sizeTabNotZero = min(fin, nbFramesTakenIntoAccount)
  
  # for i in range(0,sizeTabNotZero):
    # instSpeed2[i] = instSpeed[i]
  instMaxTailBendPos2[0:sizeTabNotZero] = instMaxTailBendPos[0:sizeTabNotZero]
  
  
  return instMaxTailBendPos2



import numpy as np
import math
import cv2

def findTheTwoSides2(headPosition, bodyContour, curFrame, hyperparameters, bestAngle):
  
  # Finding the 'mouth' of the fish
  unitVector = np.array([math.cos(bestAngle), math.sin(bestAngle)])
  factor     = 1
  headPos    = np.array(headPosition)
  testBorder = headPos + factor * unitVector
  testBorder = testBorder.astype(int)
  while (cv2.pointPolygonTest(bodyContour, (testBorder[0], testBorder[1]), True) > 0) and (factor < 100) and (testBorder[0] >= 0) and (testBorder[1] >= 0) and (testBorder[0] < len(curFrame[0])) and (testBorder[1] < len(curFrame)):
    factor = factor + 1
    testBorder = headPos + factor * unitVector
  
  # Finding the indexes of the two "border points" along the contour (these are the two points that are the closest from the 'mouth' of fish)
  xOtherBorder = testBorder[0]
  yOtherBorder = testBorder[1]
  minDist1 = 1000000000000
  minDist2 = 1000000000000
  indMin1  = 0
  indMin2  = 0
  for i in range(0, len(bodyContour)):
    Pt   = bodyContour[i][0]
    dist = math.sqrt((Pt[0] - xOtherBorder)**2 + (Pt[1] - yOtherBorder)**2)
    if (dist < minDist1):
      minDist2 = minDist1
      indMin2  = indMin1
      minDist1 = dist
      indMin1  = i
    else:
      if (dist < minDist2):
        minDist2 = dist
        indMin2  = i
  
  res = [indMin1, indMin2]
  
  return res


import h5py
import numpy as np
import cv2
import math
import json
import sys
from scipy import interpolate
from zebrazoom.code.getImage.getForegroundImage import getForegroundImage
from zebrazoom.code.getImage.headEmbededFrame import headEmbededFrame
from scipy.interpolate import UnivariateSpline
from numpy import linspace
# import time

def findTheTwoSides(headPosition, bodyContour, dst, hyperparameters):
  
  if hyperparameters["detectMouthInsteadOfHeadTwoSides"] == 1:
    
    lenX = len(dst[0])
    lenY = len(dst)
    originalShape = np.zeros((lenY, lenX))
    originalShape[:, :] = 0
    originalShape = originalShape.astype(np.uint8)
    cv2.fillPoly(originalShape, pts =[bodyContour], color=(255))
    
    # Heading calculation: first approximation
    minWhitePixel = 1000000000
    bestAngle     = 0
    nTries        = 50
    for i in range(0, nTries):
      angleOption = i * ((2 * math.pi) / nTries)
      startPoint = (int(headPosition[0]), int(headPosition[1]))
      endPoint   = (int(headPosition[0] + 100000 * math.cos(angleOption)), int(headPosition[1] + 100000 * math.sin(angleOption)))
      testImage  = originalShape.copy()
      testImage  = cv2.line(testImage, startPoint, endPoint, (0), 1)
      nbWhitePixels = cv2.countNonZero(testImage)
      if nbWhitePixels < minWhitePixel:
        minWhitePixel = nbWhitePixels
        bestAngle     = angleOption
    bestAngleAfterFirstStep = bestAngle
    
    # Heading calculation: second (and refined) approximation
    # Searching for the optimal value of iterationsForErodeImageForHeadingCalculation
    countTries = 0
    nbIterations2nbWhitePixels = {}
    if "iterationsForErodeImageForHeadingCalculation" in hyperparameters:
      iterationsForErodeImageForHeadingCalculation = hyperparameters["iterationsForErodeImageForHeadingCalculation"]
    else:
      iterationsForErodeImageForHeadingCalculation = 4
    kernel = np.ones((3, 3), np.uint8)
    nbWhitePixelsMax = 75
    while (iterationsForErodeImageForHeadingCalculation > 0) and (countTries < 50) and not(iterationsForErodeImageForHeadingCalculation in nbIterations2nbWhitePixels):
      testImage2 = cv2.erode(testImage, kernel, iterations = iterationsForErodeImageForHeadingCalculation)
      nbWhitePixels = cv2.countNonZero(testImage2)
      nbIterations2nbWhitePixels[iterationsForErodeImageForHeadingCalculation] = nbWhitePixels
      if nbWhitePixels < nbWhitePixelsMax:
        iterationsForErodeImageForHeadingCalculation = iterationsForErodeImageForHeadingCalculation - 1
      if nbWhitePixels >= nbWhitePixelsMax:
        iterationsForErodeImageForHeadingCalculation = iterationsForErodeImageForHeadingCalculation + 1
      countTries = countTries + 1
    best_iterations = 0
    minDist = 10000000000000
    for iterations in nbIterations2nbWhitePixels:
      nbWhitePixels = nbIterations2nbWhitePixels[iterations]
      dist = abs(nbWhitePixels - nbWhitePixelsMax)
      if dist < minDist:
        minDist = dist
        best_iterations = iterations
    iterationsForErodeImageForHeadingCalculation = best_iterations
    hyperparameters["iterationsForErodeImageForHeadingCalculation"] = iterationsForErodeImageForHeadingCalculation
    
    maxDist = -1
    for i in range(0, nTries):
      angleOption = bestAngleAfterFirstStep - (math.pi / 5) + i * ((2 * (math.pi / 5)) / nTries)

      startPoint = (int(headPosition[0]), int(headPosition[1]))
      endPoint   = (int(headPosition[0] + 100000 * math.cos(angleOption)), int(headPosition[1] + 100000 * math.sin(angleOption)))
      testImage  = originalShape.copy()
      
      # applying dilation with the iterationsForErodeImageForHeadingCalculation value found
      testImage = cv2.erode(testImage, kernel, iterations = iterationsForErodeImageForHeadingCalculation)
      testImage  = cv2.line(testImage, startPoint, endPoint, (0), 1)
      nbWhitePixels = cv2.countNonZero(testImage)
      if nbWhitePixels < minWhitePixel:
        minWhitePixel = nbWhitePixels
        bestAngle     = angleOption
    
    # Finding the 'mouth' of the fish
    unitVector = np.array([math.cos(bestAngle + math.pi), math.sin(bestAngle + math.pi)])
    factor     = 1
    headPos    = np.array(headPosition)
    testBorder = headPos + factor * unitVector
    testBorder = testBorder.astype(int)
    while (cv2.pointPolygonTest(bodyContour, (testBorder[0], testBorder[1]), True) > 0) and (factor < 100) and (testBorder[0] >= 0) and (testBorder[1] >= 0) and (testBorder[0] < len(dst[0])) and (testBorder[1] < len(dst)):
      factor = factor + 1
      testBorder = headPos + factor * unitVector
    
    # Finding the indexes of the two "border points" along the contour (these are the two points that are the closest from the 'mouth' of fish)
    xOtherBorder = testBorder[0]
    yOtherBorder = testBorder[1]
    minDist1 = 1000000000000
    minDist2 = 1000000000000
    indMin1  = 0
    indMin2  = 0
    for i in range(0, len(bodyContour)):
      Pt   = bodyContour[i][0]
      dist = math.sqrt((Pt[0] - xOtherBorder)**2 + (Pt[1] - yOtherBorder)**2)
      if (dist < minDist1):
        minDist2 = minDist1
        indMin2  = indMin1
        minDist1 = dist
        indMin1  = i
      else:
        if (dist < minDist2):
          minDist2 = dist
          indMin2  = i
      
    res = [indMin1, indMin2, bestAngle + math.pi]
    
  else:
    
    res = np.zeros(2)
    
    x = headPosition[0]
    y = headPosition[1]
    
    minDist = 1000000000000
    indMin  = 0
    for i in range(0, len(bodyContour)):
      Pt   = bodyContour[i][0]
      dist = math.sqrt((Pt[0] - x)**2 + (Pt[1] - y)**2)
      if (dist < minDist):
        minDist = dist
        indMin  = i
    
    res[0] = indMin
    PtClosest = bodyContour[indMin][0]
    headPos   = np.array(headPosition)
    
    unitVector = np.array([x - PtClosest[0], y - PtClosest[1]])
    unitVectorLength = math.sqrt(unitVector[0]**2 + unitVector[1]**2)
    unitVector[0] = unitVector[0] / unitVectorLength
    unitVector[1] = unitVector[1] / unitVectorLength
    
    factor = 1
    testBorder = headPos + factor * unitVector
    testBorder = testBorder.astype(int)
    while (cv2.pointPolygonTest(bodyContour, (testBorder[0], testBorder[1]), True) > 0) and (factor < 100) and (testBorder[0] >= 0) and (testBorder[1] >= 0) and (testBorder[0] < len(dst[0])) and (testBorder[1] < len(dst)):
      factor = factor + 1
      testBorder = headPos + factor * unitVector
    
    xOtherBorder = testBorder[0]
    yOtherBorder = testBorder[1]
    
    minDist = 1000000000000
    indMin2  = 0
    for i in range(0, len(bodyContour)):
      Pt   = bodyContour[i][0]
      dist = math.sqrt((Pt[0] - xOtherBorder)**2 + (Pt[1] - yOtherBorder)**2)
      if (dist < minDist):
        minDist = dist
        indMin2  = i
    
    res[1] = indMin2
    
  if False:
    import zebrazoom.code.util as util

    cv2.circle(dst, (pt1[0],pt1[1]), 1, (0, 0, 255), -1)
    cv2.circle(dst, (pt2[0],pt2[1]), 1, (0, 0, 255), -1)
    util.showFrame(dst, title='Frame')
  
  return res



import os

from zebrazoom.extractZZParametersFromTailAngle import extractZZParametersFromTailAngle
from zebrazoom.dataAnalysis.datasetcreation.createDataFrame import createDataFrame
from zebrazoom.dataAnalysis.dataanalysis.populationComparaison import populationComparaison
from zebrazoom.dataAnalysis.dataanalysis.applyClustering import applyClustering


with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'version.txt'), 'r') as f:
  __version__ = f.read().strip()
